from secrets import randbits, token_bytes
#from  os.path import join
from pickle import dumps, loads
#from copy import deepcopy
import time


#from typing import Union, NoReturn, TypeVar, List

#from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
#from cryptography.hazmat.primitives.asymmetric.x448 import X448PrivateKey, X448PublicKey

#from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey
#from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PrivateKey, Ed448PublicKey
#from cryptography.hazmat.primitives.asymmetric import rsa
#from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
#from cryptography.hazmat.primitives.asymmetric import ec
#from cryptography.hazmat.primitives.asymmetric.ec import
# SECP256R1, SECP384R1, SECP521R1, ECDSA, EllipticCurvePublicNumbers, ECDH
#, EllipticCurvePrivateKey, EllipticCurvePublicKey

from cryptography.hazmat.primitives.hashes import Hash, SHA256, SHA384, SHA512
#from cryptography.hazmat.primitives.serialization import load_der_private_key, 
#load_pem_private_key, NoEncryption, Encoding, PrivateFormat, PublicFormat
#from cryptography import x509
#from cryptography.x509.oid import NameOID

#from cryptography.hazmat.primitives.asymmetric import padding
#from cryptography.hazmat.primitives import serialization
#from cryptography.hazmat.primitives.kdf.hkdf import HKDF, HKDFExpand
#from cryptography.hazmat.primitives.hmac import HMAC

#import datetime
#import hashlib

#from construct.core import *
#from construct.lib import *

import pylurk.debug
#import pylurk.tls13.struct_tls13 as lurk
#from pylurk.debug import get_struct, get_struct_index
from pylurk.tls13.key_scheduler import KeyScheduler
from pylurk.tls13.tls_handshake import TlsHandshake
import pylurk.tls13.crypto_suites

import pylurk.conf
from pylurk.tls13.crypto_suites import SigScheme, CipherSuite
from pylurk.lurk.lurk_lurk import LURKError, ImplementationError
#, ConfigurationError
import pylurk


class Tag:

  def __init__( self, tag:dict, mtype, tls13_conf, ctx=None ):
    self.conf = tls13_conf
    self.mtype = mtype
    self.last_exchange = tag[ 'last_exchange' ]

    if self.mtype in [ 's_init_cert_verify', 's_hand_and_app_secret'  ]:
      self.last_exchange = self.conf[ 'last_exchange' ][ self.mtype ]
    elif self.mtype in [ 's_new_ticket', 'c_register_tickets' ]:
      if ctx >= self.conf[ 'max_tickets' ]:
        self.last_exchange = True
    elif self.mtype in [ 'c_init_client_finished' ]:
      ## ctx is the handshake
      self.last_exchange = self.conf[ 'last_exchange' ][ self.mtype ]
      if ctx.is_post_hand_auth_proposed() is False:
        self.last_exchange = True
    elif self.mtype in [ 'c_post_hand_auth' ]:
      if ctx >= self.conf[ 'max_post_handshake_authentication' ]:
        self.last_exchange = True
    elif self.mtype in [ 'c_client_finished' ]:
      ## ctx is the handshake
      self.last_exchange = self.conf[ 'last_exchange' ][ self.mtype ]
      if ctx.is_post_hand_auth_proposed() is False and\
         ctx.msg_list[ 0 ][ 'msg_type' ] == 'server_hello' :
        self.last_exchange = True
    else:
      raise ImplementationError( f"unknown type {self.mtype}" )

    self.resp = { 'last_exchange' : self.last_exchange }


class Ephemeral:

  def __init__(self, ephemeral:dict, mtype, tls13_conf, handshake=None,\
               client_hello_ephemeral=None, debug=None ) -> None:
    """ initializes the object based on the structure """

    self.conf = tls13_conf
    self.ephemeral = ephemeral
    self.mtype = mtype
    self.debug = debug
    ## The s_init_early_data only requires to store the cliemt_shares
    if ephemeral != {}:
      self.method = self.ephemeral['method']
    self.handshake = handshake
    self.sanity_check( )
    ## the entry for the key scheduler
    if self.conf[ 'role' ] == 'server':
      if self.mtype == 's_init_cert_verify':
        self.shared_secret, self.resp = self.server_shared_secret( )
        self.server_share = self.resp[ 'key' ]
      if self.mtype == 's_init_early_secret' :
        #self.client_ecdhe_key_list = self.get_key_share_client_shares()
        self.client_ecdhe_key_list = self.client_shares()
      elif self.mtype == 's_hand_and_app_secret':
        self.shared_secret, self.resp = self.server_shared_secret( client_hello_ephemeral )
        self.server_share = self.resp[ 'key' ]
      ## key is the server_share value which is a key share entry
    elif self.conf[ 'role' ] == 'client':
      if self.mtype == 'c_init_client_hello':
        self.client_ecdhe_key_list, self.resp = self.client_shares( )
      if self.mtype == 'c_init_client_finished':
        if self.method == 'e_generated':
          self.shared_secret = self.ephemeral[ 'key' ][ 'shared_secret' ]
        elif self.method == 'no_secret':
          self.shared_secret = None
    else:
      raise ImplementationError( f"unknown role {self.conf[ 'role' ]}" )

    ## attributes defined latter
    self.server_ecdhe_key = None


  def sanity_check( self ):
    """ check coherence of ephemeral with mtype and handshake """
    if self.mtype == 's_init_early_secret' and self.ephemeral == {}:
      pass
    elif self.method not in self.conf['ephemeral_method_list']:
      raise LURKError( 'invalid_ephemeral', f"method {self.method} expected to be"\
                       "in {self.conf['ephemeral_method_list']}" )
    if ( self.mtype == 's_init_cert_verify' and self.method == 'no_secret' ) or\
       ( self.mtype == 'c_client_finished' and self.method == 'cs_generated' ) :
      raise LURKError( 'invalid_ephemeral',\
              f"Incompatible {self.method} and {self.mtype}" )
    elif ( self.mtype == 's_hand_and_app_secret' and self.method == 'no_secret' ):
      if self.handshake.is_ks_agreed() :
        raise LURKError( 'invalid_ephemeral', \
                "unexpected key_share extension with 'no_secret'" )

    elif self.mtype == 'c_init_client_finished':
      if self.method == 'cs_generated' :
        raise LURKError( 'invalid_ephemeral', \
                f"Incompatible {self.method} and {self.mtype}" )
      if self.method == 'no_secret' and\
              not ( self.handshake.is_psk_proposed() \
              and self.handshake.is_psk_agreed() ) :
        raise LURKError( 'invalid_ephemeral', f"no (EC)DHE provided ({self.method})"\
                f"but PSK (without (EC)DHE) authentication is not agreed" )
      if self.method == 'e_generated' and not self.handshake.is_ks_agreed() :
        raise LURKError( 'invalid_ephemeral', \
                f"(EC)DHE provided ({self.method})"\
                f"but PSK-ECDHE or ECDHE authentication is not agreed" )
    elif self.mtype == 'c_init_client_hello':
      pass
    elif self.mtype == 'c_server_hello':
      if self.method in [ 'cs_generated', 'e_generated' ] and not self.handshake.is_ks_agreed() :
        raise LURKError( 'invalid_ephemeral', \
                f"(EC)DHE provided ({self.method}) "\
                f"but PSK-ECDHE or ECDHE authentication is not agreed "\
                f"{self.handshake.msg_list}" )
      elif self.method == 'no_secret' and self.handshake.is_ks_agreed() :
        raise LURKError( 'invalid_ephemeral', \
                f"(EC)DHE not provided ({self.method})"\
                f"but PSK-ECDHE or ECDHE authentication is agreed" )

  def proceed_empty_key_share_entry( self, empty_ks_entry ):
    """ processes an  empty key share entry

    Returns:
      - private key
      - key share entry
    """
    group = empty_ks_entry[ 'group' ]
    key_exchange = empty_ks_entry[ 'key_exchange' ]
    if group not in self.conf[ 'authorized_ecdhe_group' ]:
      raise LURKError( 'invalid_ephemeral', f"unsupported {group}" )
    if key_exchange not in [ b'', None]:
      raise LURKError( 'invalid_ephemeral', \
              f"expecting empty key share entry {empty_ks_entry}" )
###
    ecdhe_key = pylurk.tls13.crypto_suites.ECDHEKey()
    ecdhe_key.group = group
    ## try to read value from test_vector file
    if self.debug is not None and self.debug.test_vector is True:
      key =f"{self.conf[ 'role' ]}_{group}_ecdhe_private"
      if key in self.debug.db.keys():
        ecdhe_key.generate_from_pem( self.debug.read_bin( key ) )
      if self.debug.check is True:
        self.debug.check_bin( ecdhe_key.pkcs8(), self.debug.read_bin( key ) )
    elif ecdhe_key.private is None and ecdhe_key.public is None:
      ecdhe_key.generate( group )
    return ecdhe_key

  def client_shares( self ):
    """ computes client_shares, private_keys and resp """
    client_shares = self.handshake.get_key_share( side='client' )
    ecdhe_key_list = []
    resp = []
    for ks in client_shares :
      if ks[ 'key_exchange' ] in [ None, b'' ]:
        ecdhe_key = self.proceed_empty_key_share_entry( ks )
        resp.append( { 'method' : 'cs_generated',
                          'key' : ecdhe_key.ks_entry() } )
      else:
        ecdhe_key = pylurk.tls13.crypto_suites.ECDHEKey( )
        ecdhe_key.generate_from_ks_entry( ks )
        resp.append( { 'method' : 'e_generated',
                            'key' : b'' } )
    ecdhe_key_list.append( ecdhe_key )
    return ecdhe_key_list, resp

  def client_shared_secret( self, c_init_client_hello_ephemeral ):
    """ compute the share secret on the client side

    The current object is generated from the ephemeral provided by E during the
    c_server_hello exchange.
    On the other hand private keys have been generated during the
    c_init_client_hello exchange.
    This latest ephemeral is used as th einput.
    """

    if self.method == 'e_generated':
      self.shared_secret = self.ephemeral[ 'key' ][ 'shared_secret' ]
    elif self.method == 'cs_generated':
      self.server_ecdhe_key = pylurk.tls13.crypto_suites.ECDHEKey( )
      self.server_ecdhe_key.generate_from_ks_entry( self.handshake.get_key_share( side='server' ) )
      for key in c_init_client_hello_ephemeral.client_ecdhe_key_list :
        if self.server_ecdhe_key.group == key.group :
          self.shared_secret = self.server_ecdhe_key.shared_secret( key )
          break
    elif self.method == 'no_secret':
      self.shared_secret = None
    else:
      raise LURKError( 'invalid_ephemeral', f"Unexpected method {self.method}" )

  def server_shared_secret( self, client_hello_ephemeral=None ):
    """ treat ephemeral extension and initializes self.ecdhe, self.server_key_exchange

    This function is responsible for generating the server (EC)DHE public key,
    computing the shared secret as well as generating the key_share extension
    of the ServerHello.
    self.shared_secret contains the shared secret key, the server (EC)DHE public key
    is taken from the reurned resp and stored in self.server_share

    see __init__( ) fucntion:
      self.shared_secret, self.resp = self.server_shared_secret( )
      self.server_share = self.resp[ 'key' ]
    """
    ## with method 'e_generated' the shared_secret is explictly provided
    if self.method == 'e_generated':
      self.shared_secret = self.ephemeral['key'][ 'shared_secret' ]
      resp = { 'method' : self.method,
               'key' : b'' }
    ## with method 'cs_generated' the cs needs to generate the public key,
    ## private key, take the client public key and compute the shared secret
    elif self.method == 'cs_generated':
      ## retrieve the server key_share entry and seklect the client key_share entry
      ## since method is 'cs_generated' the server key_share entry MUST be empty
      self.server_ecdhe_key = pylurk.tls13.crypto_suites.ECDHEKey()
      server_ks = self.handshake.get_key_share( side='server' )
      self.server_ecdhe_key.generate( group=server_ks[ 'group' ] )

      if client_hello_ephemeral is None:
        self.client_ecdhe_key_list, tmp_resp = self.client_shares()
      else:
        self.client_ecdhe_key_list = client_hello_ephemeral.client_ecdhe_key_list
      for key in self.client_ecdhe_key_list:
        if key.group == self.server_ecdhe_key.group :
          self.shared_secret = key.shared_secret( self.server_ecdhe_key )
          break
      resp = { 'method' : self.method,
               'key' : self.server_ecdhe_key.ks_entry()  }
    elif self.method == 'no_secret':
      self.shared_secret = None
      resp = { 'method' : self.method,
               'key' : b'' }
    else:
      raise LURKError( 'invalid_ephemeral', f"Unexpected method {self.method}" )
    return self.shared_secret, resp

class SessionID:
  def __init__( self, session_id:bytes, tag=None ):
    if tag is None :
      self.cs = token_bytes( 4  )
      self.e = session_id
    else:
      if tag.last_exchange is False:
        self.cs = token_bytes( 4  )
        self.e = session_id
      else:
        self.cs = b''

  def is_in_session( self, mtype, status, session_id:bytes ):
    """ checks the session_id

    Checks the value is aligned with the sessionID object of the session
    """

    if ( status == 'request' and 'init' in mtype and session_id == self.e ) or \
       ( status == 'request' and 'init' not in mtype and session_id == self.cs ) or \
       ( status != 'request' and 'init' in mtype and session_id == self.cs ) or \
       ( status != 'request' and 'init' not in mtype and session_id == self.e ) :
      return True
    raise LURKError( 'invalid_session_id', f"{mtype} - E ( {self.e} / CS {self.cs}" )


class Freshness:
  def __init__( self, freshness:str ):
    self.freshness_funct = freshness

  def update_random( self, role, random:bytes ):
    if self.freshness_funct == 'null' :
      return random
    if role == 'server':
      ctx = b"tls13 pfs srv"
    elif role == 'client':
      ctx = b"tls13 pfs clt"
    if self.freshness_funct == 'sha256' :
      digest = Hash( SHA256() )
    elif self.freshness_funct == 'sha384' :
      digest = Hash( SHA384() )
    elif self.freshness_funct == 'sha512' :
      digest = Hash( SHA512() )
    else:
      raise LURKError( 'invalid_freshness', f"{self.freshness_funct}" )
    ## random does not include ctx
    ## this needs to be changed
    digest.update( random )
    return digest.finalize()

class LurkCert:

  def __init__( self, lurk_cert:dict, mtype, tls13_conf, server, handshake=None ):
    """ Handles the treatment of LurkCert

    Args:
      server (bool) indicates the nature of the certificate.
        When

    """
    self.conf = tls13_conf
    self.handshake = handshake
    self.cert_type = lurk_cert[ 'cert_type' ]
    self.uncompress_lurk_cert( lurk_cert )
    self.sanity_check( mtype, server )

  def uncompress_lurk_cert( self, lurk_cert ):
    """ decompress lurk_cert into a TLS 1.3 Certificate structure  """
    if self.cert_type == 'no_certificate' :
      self.certificate = None
    elif self.cert_type == 'uncompressed' :
      self.certificate = lurk_cert[ 'certificate' ]
    elif self.cert_type == 'compressed' :
      self.certificate = 'XXX'
    elif self.cert_type == 'finger_print' :
      finger_print_dict = self.conf[ '_finger_print_dict' ]
      cert_entry_list = []
      finger_print_entry_list = lurk_cert[ 'certificate' ][ 'certificate_list' ]
      try:
        for entry in finger_print_entry_list :
          cert_entry = { 'cert' : finger_print_dict[ entry[ 'finger_print' ] ], \
                         'extensions' : entry[ 'extensions' ][ : ] }
          cert_entry_list.append( cert_entry )
      except KeyError:
        raise LURKError( 'invalid_certificate', f"unrecognized fingerprint "\
                         f"{finger_print_entry_list} {finger_print_dict}" )
      else:
        if self.conf[ 'role' ] == 'server':
          cert_req_ctx = b''
        elif self.conf[ 'role' ] == 'client':
          cert_req_index = self.handshake.msg_type_list().index( 'certificate_request' )
          cert_req_ctx = self.handshake.msg_list[ cert_req_index ]\
                            [ 'data' ][ 'certificate_request_context' ]
        else:
          raise ImplementationError( f"unknown role {self.conf[ 'role' ]}" )
        self.certificate = { 'certificate_request_context' : cert_req_ctx, \
                        'certificate_list' : cert_entry_list }
    else:
      raise LURKError( 'invalid_certificate', "unknown cert_type {self.cert_type}" )
    if  self.certificate is not None:
      self.hs_cert_msg = { 'msg_type' : 'certificate',
                           'data' : self.certificate }
    else:
      self.hs_cert_msg = {}

  def sanity_check( self, mtype, server:bool ):
    if mtype == 's_new_ticket' :
      if self.cert_type == 'no_certificate' :
        if 'certificate_verify' in self.handshake.msg_type_list() :
          raise LURKError( 'invalid_certificate', f"Incompatible server " \
            f"cert_type {self.cert_type} with handshake. Unexpected "\
            f"CertificateVerify message - {self.handshake.msg_type_list()}" )
      else :
        if 'certificate_verify' not in self.handshake.msg_type_list() :
          raise LURKError( 'invalid_certificate', f"Incompatible server " \
            f"cert_type {self.cert_type} with handshake. Expected "\
            f"CertificateVerify message - {self.handshake.msg_type_list()}" )
    elif mtype == 'c_init_client_finished' :
      ## server certificate
      if server is True:
        if self.cert_type == 'no_certificate' :
          if not self.handshake.is_psk_proposed()  or \
             not self.handshake.is_psk_agreed() :
            raise LURKError( 'invalid_certificate', f"Incompatible server " \
              f"cert_type {self.cert_type} with handshake. Expecting PSK "\
              f"authentication" )
        else:
          if not self.handshake.is_certificate_agreed() :
            raise LURKError( 'invalid_certificate', f"Incompatible server " \
              f"cert_type {self.cert_type} with handshake. Expecting certificate "\
              f"authentication" )
      ## client certificate
      else :
        if  ( (self.cert_type != 'no_certificate' ) and \
            self.handshake.is_certificate_request() ) is False:
          raise LURKError( 'invalid_certificate', \
                  "Client certificate and CertificateRequest " +\
                  "MUST either be together absent or present" )

class SecretReq:

#  def __init__( self, secret_request:dict, mtype,\
#        tls13_conf, handshake=None ):
  def __init__( self, secret_request:dict, mtype, tls13_conf ):

    self.conf = tls13_conf
    if mtype in [ 's_init_early_secret', 'c_init_client_hello' ]:
      mandatory = ['b' ]
      forbiden = [ 'h_c', 'h_s', 'a_c', 'a_s', 'x', 'r' ]
      optional = [ 'e_s', 'e_x' ]
    elif mtype in [ 's_init_cert_verify', 's_hand_and_app_secret' ]:
      mandatory = [ 'h_c', 'h_s' ]
      forbiden = [ 'b', 'e_s', 'e_x', 'r' ]
      optional = [ 'a_c', 'a_s', 'x' ]
    elif mtype in [ 's_new_ticket', 'c_register_ticket' ]:
      mandatory = []
      forbiden = [ 'b', 'e_s', 'e_x', 'h_c', 'h_s', 'a_c', 'a_s', 'x' ]
      optional = ['r']
    elif mtype in [ 'c_server_hello' ]:
      mandatory = [ 'h_c', 'h_s' ]
      forbiden = [ 'b', 'e_s', 'e_x', 'a_c', 'a_s', 'x', 'r' ]
      optional = []
    elif mtype in [ 'c_client_finished' ]:
      mandatory = [ 'a_c', 'a_s' ]
      forbiden = [ 'b', 'e_s', 'e_x', 'a_c', 'a_s', 'x' ]
      optional = [ 'x', 'r' ]
    else:
      raise ImplementationError( f"unknown {mtype}" )

    self.authorized_secrets = mandatory
    ## building the list of requested secrets, that is
    ## those set to True in key_request
    for key in secret_request.keys():
      if secret_request[ key ] is True and key not in forbiden :
        self.authorized_secrets.append( key )
    self.authorized_secrets = list( set( self.authorized_secrets ) )

    if mtype in [ 's_init_cert_verify', 'c_hand_and_app_secret' ] :
      if self.conf[ 'app_secret_authorized' ] is False :
        try:
          self.authorized_secrets.remove( 'a_c' )
          self.authorized_secrets.remove( 'a_s' )
        except KeyError:
          pass
      if self.conf[ 'exporter_secret_authorized' ] is False:
        try:
          self.authorized_secrets.remove( 'x' )
        except KeyError:
          pass
    elif mtype == 's_new_ticket':
      if self.conf[ 'resumption_secret_authorized' ] is False:
        try:
          self.authorized_secrets.remove( 'r' )
        except KeyError:
          pass
    elif mtype in [ 's_init_early_secret', 'c_init_client_hello' ]:
      if self.conf[ 'client_early_secret_authorized' ] is False:
        try:
          self.authorized_secrets.remove( 'e_s' )
        except KeyError:
          pass
      if self.conf[ 'early_exporter_secret_authorized' ] is False:
        try:
          self.authorized_secrets.remove( 'e_x' )
        except KeyError:
          pass

  def of( self, secret_candidates:list)-> list:
    """returns the sublist of secrets both authorized and requested.

     Usually useful to determine which secrets need to be computed """
    secret_list = []
    for secret in secret_candidates:
      if secret in self.authorized_secrets:
        secret_list.append( secret )
    return secret_list

  def resp( self, scheduler ):
    secret_list = []
    for secret_type in self.authorized_secrets:
      secret_list.append( \
        { 'secret_type' : secret_type,
          'secret_data' : scheduler.secrets[ secret_type ] } )
    return secret_list

class SessionTicket:
  """ TLS Session Tickets for session resumption

  Attributes:
    conf: a dictionary containing the TLS 1.3 configuration 
      parameters
    psk_identity: the identity of the PSK expressed in bytes
    cipher: the TLS cipher suite
    tls_hash: The TLS hash function. Currenlty its is directly genberated from the cipher suite.
    psk: The PSK 

  """
## We need to be able to generate encrypted tickets.
  def __init__( self, tls13_conf:dict, psk_identity:bin = None ):
    """ session ticket

      psk_identity (bin):
    """
    self.conf = tls13_conf
    self.psk_identity = psk_identity
    self.cipher = None
    self.tls_hash = None
    self.psk = None

    if psk_identity is not None:
      self.read_ticket( psk_identity )
    ## This variable is used to identify the selection of the currenlty
    ## used ticket.
    ## This is used to check the server hello match the ticket that has
    ## been used to generates the early secrets
    self.selected_identity = None


  def new( self, scheduler, cipher:str ) -> dict :
    """ creates a new session ticket

    Args:
      scheduler: the scheduler object from which the psk will
        be generated
      cipher: the TLS cipher associated to the Ticket Session
        will be generated
    Returns:
      the structure of a New Session Ticket
    """
    self.cipher = cipher
    self.tls_hash = CipherSuite( self.cipher ).get_hash()
    ticket_nonce = token_bytes( self.conf[ 'ticket_nonce_len' ] )
    self.psk =  scheduler.compute_psk( ticket_nonce )

    ticket = dumps( { 'cipher' : cipher, 'psk': self.psk } )
    return { \
      'ticket_lifetime' : self.conf[ 'ticket_life_time' ], \
      'ticket_age_add' : randbits( 4 * 8 ), \
      'ticket_nonce' : ticket_nonce, \
      'ticket' : ticket,
      'extensions' :[] }

  def read_ticket( self, psk_identity:dict ):

    ticket =  loads( psk_identity[ 'identity' ] )
    self.cipher = ticket[ 'cipher' ]
    self.psk = ticket[ 'psk' ]
    self.tls_hash = CipherSuite( self.cipher ).get_hash()

  def init_handshake( self, handshake ):
    """initialize handshake from values stored in the ticket.

    This is needed for example in the abscente of a server hello when
    the hash function needs to be known to generates the early secrets.
    The hash function is in fact generated from the cipher suite, so the
    hash is not directly configured.
    see SInitEarlySecret for an example.
    """
    handshake.cipher_suite = self.cipher

class SInitCertVerifyReq:

  def __init__(self, req, tls13_conf, debug=None ):
    self.conf = tls13_conf
    self.mtype = 's_init_cert_verify'
    self.next_mtype = 's_new_ticket'

    self.freshness = Freshness( req[ 'freshness' ] )
    self.secret_request = SecretReq(req[ 'secret_request' ], self.mtype, self.conf )
    self.sig_algo = SigScheme( req[ 'sig_algo' ] )
    self.handshake = TlsHandshake( 'server', self.conf )
    self.handshake.msg_list.extend( req[ 'handshake' ] )
    self.handshake.sanity_check( self.mtype )
    self.handshake.update_random( self.freshness )
    self.cert = LurkCert( req[ 'certificate' ], self.mtype, self.conf, True, \
                          self.handshake )
    self.ephemeral = Ephemeral( req[ 'ephemeral' ], self.mtype, self.conf, self.handshake )
    self.scheduler = None

    if self.ephemeral.method == 'cs_generated':
      self.handshake.update_key_share( self.ephemeral.server_share )
    self.scheduler = KeyScheduler( self.handshake.get_tls_hash(), \
                                   shared_secret=self.ephemeral.shared_secret )
    self.scheduler.process( self.secret_request.of([ 'h_c', 'h_s' ] ), self.handshake )
    self.handshake.update_certificate( self.cert )
    self.handshake.update_certificate_verify( )
    ## get sig from freshly generated certificate_verify
    sig = self.handshake.msg_list[ 0 ][ 'data' ]['signature' ]
#    self.handshake.update_server_finished( self.scheduler )
    self.handshake.update_finished( self.scheduler )
    self.scheduler.process( self.secret_request.of( [ 'a_c', 'a_s', 'x' ] ), self.handshake )
    self.tag = Tag( req[ 'tag' ], self.mtype, self.conf )
    self.session_id = SessionID( req[ 'session_id' ], self.tag )
    self.resp = { 'tag' : self.tag.resp,
                  'session_id' : self.session_id.cs,
                  'ephemeral' : self.ephemeral.resp,
                  'secret_list' : self.secret_request.resp( self.scheduler ),
                  'signature' : sig }

class SNewTicketReq:

  def __init__( self, req, tls13_conf, handshake, scheduler, session_id, \
                ticket_counter, debug=None ):
    self.conf = tls13_conf
    self.mtype = 's_new_ticket'
    self.handshake = handshake
    self.scheduler = scheduler
    self.ticket_counter = ticket_counter

    self.handshake.msg_list.extend( req[ 'handshake' ] )
    self.cert = LurkCert( req[ 'certificate' ], self.mtype, self.conf, False, \
                          self.handshake )
    if self.cert.cert_type != 'no_certificate' :
      self.handshake.update_certificate( self.cert, server=False )
    self.secret_request = SecretReq(req[ 'secret_request' ], self.mtype, self.conf )
    self.scheduler.process( self.secret_request.of( [ 'r' ] ), self.handshake )
    self.next_mtype = 's_new_ticket'

    ticket_nbr = self.nbr( req[ 'ticket_nbr' ] )
    ticket_list = []
    ticket = SessionTicket( self.conf )
    for t in range( ticket_nbr ):
      ticket_list.append( \
        ticket.new( self.scheduler, self.handshake.get_cipher_suite() ) )
    self.tag = Tag( req[ 'tag' ], self.mtype, self.conf, ctx=ticket_nbr )

    self.resp =  { 'tag' : self.tag.resp,
              'session_id' : session_id.e,
              'secret_list' : self.secret_request.resp( self.scheduler ),
              'ticket_list' : ticket_list }


  def nbr( self, ticket_nbr ):
    """ determine the maximum number of tickets that can be emitted

    use to compute self.ticket_nbr
    """
    if self.ticket_counter <= self.conf[ 'max_tickets' ]:
      n =  min( ticket_nbr, self.conf[ 'max_tickets' ] )
      self.ticket_counter += n
    else:
      n = 0
    return n

class SInitEarlySecretReq:

  def __init__( self, req, tls13_conf, debug=None ):
    self.conf = tls13_conf
    self.mtype = 's_init_early_secret'
    self.next_mtype = 's_hand_and_app_secret'

    self.freshness = Freshness( req[ 'freshness' ] )

    self.handshake = TlsHandshake( 'server', self.conf )
    self.handshake.msg_list.extend( req[ 'handshake' ] )
    self.handshake.sanity_check( self.mtype )
    if self.handshake.is_early_data_proposed( ) is False :
      req[ 'secret_request' ][ 'e_s' ] = False
    ## the binary format of the ticket
    psk_identity = self.handshake.get_ticket( req[ 'selected_identity' ] )
    self.session_ticket = SessionTicket( self.conf, psk_identity=psk_identity )
    ## to enable to check the server hello is conformed with the ticket in used.
    self.session_ticket.selected_identity = req[ 'selected_identity' ]
    self.session_ticket.init_handshake( self.handshake )
    self.secret_request = SecretReq(req[ 'secret_request' ], \
                          self.mtype, self.conf )
    ## only to store the client ephemeral that will be used in
    ## SHandAndAppSecret exchange
    self.ephemeral = None
    if self.handshake.is_ks_proposed( ) is True:
      self.ephemeral = Ephemeral( {}, self.mtype, self.conf, self.handshake )
    self.scheduler = KeyScheduler( self.session_ticket.tls_hash, \
                                   psk=self.session_ticket.psk, is_ext=False, debug=debug )
#    self.last_exchange = None
    self.session_id = SessionID( req[ 'session_id' ] )
    self.scheduler.process( self.secret_request.of([ 'b', 'e_s', 'e_x' ] ), self.handshake )
    self.resp = { 'session_id' : self.session_id.cs,
             'secret_list' : self.secret_request.resp( self.scheduler ) }

class SHandAndAppSecretReq:

  def __init__( self, req, tls13_conf, handshake, scheduler, session_id,\
                session_ticket, freshness, client_hello_ephemeral, debug=None ):
    self.conf = tls13_conf
    self.mtype = 's_hand_and_app_secret'
    self.next_mtype = 's_new_ticket'
    self.freshness = freshness
    self.handshake = handshake
    self.secret_request = SecretReq(req[ 'secret_request' ], self.mtype, self.conf )
    self.handshake.msg_list.extend( req[ 'handshake' ] )
    self.scheduler = scheduler
    self.session_ticket = session_ticket
    self.handshake.sanity_check( self.mtype, session_ticket=session_ticket )
    self.handshake.update_random( self.freshness )
    self.ephemeral = Ephemeral( req[ 'ephemeral' ], self.mtype, self.conf,\
                                self.handshake, client_hello_ephemeral )

    if self.ephemeral.method == 'cs_generated':
      self.handshake.update_key_share( self.ephemeral.server_share )
    self.scheduler.shared_secret = self.ephemeral.shared_secret
    self.scheduler.process( self.secret_request.of( [ 'h_c', 'h_s'] ), self.handshake )
    self.handshake.update_server_finished( self.scheduler )
    self.scheduler.process( self.secret_request.of( [ 'a_c', 'a_s', 'x' ] ), self.handshake )
    self.tag = Tag( req[ 'tag' ], self.mtype, self.conf )
    self.resp =  { 'tag' : self.tag.resp,
                   'session_id' : session_id.e,
                   'ephemeral' : self.ephemeral.resp,
                   'secret_list' : self.secret_request.resp( self.scheduler ) }

class SSession:
  """Handles a LURK session beetween the LURK client and CS

  A session is defined by the LURK messages  exchanges between
  the LURK client and the CS for setting a given TLS exchange.
  These LURK Messages may share a specific context and at least
  follows a deterministic order. 
  This object guarantee that specific order is respected and 
  that the appropriated context is shared between the messages.  

  SSession handles LURK sessions between in the TLS server

  Attributes:
    conf: a dictionnary characterizing the configuration of the CS
    next_mtype: the designation of teh next LURK message 
    scheduler: the object implementing the sheduler of the ongoing
      TLS session
    handshake: the object that characterizes the ongoing TLS
      handshake 
    session_id: the identifier of the LURK session between the
      LURK client and the CS  
    freshness: the algorithm from which randoms are generated.
    ephemeral: the ephemeral object provided by the LURK client.
    last_exchange: determine if the curent exchange between 
      the LURK client and teh CS is the last one - in which case
      the context will be removed, or if further exchanges
      are expected.
    ticket_counter: the number of ticket that have been sent
    session_ticket: the current session ticket object
    session_db: the database that contains all sessions. This
      is where that session context will be stored / retrieved.
    ticket_db: the database that contains the ticket information.
    debug: a boolean value that indicates information needs to 
      be output or not.
  """
  def __init__( self, tls13_conf:dict, session_db=None, ticket_db=None, debug=None ): # type Conf
    """ handles the various requests associated to a given session """
    self.conf = tls13_conf
    self.next_mtype = None
    self.scheduler = None
    self.handshake = None
    self.session_id = None
    self.freshness = None
    self.ephemeral = None
#    self.next_mtype = None
    self.last_exchange = None
    self.ticket_counter = 0
    self.session_ticket = None
    self.session_db = session_db
    self.ticket_db = ticket_db
    self.debug=debug

  def is_expected_message( self, mtype, status ):
    """ check the message type is expected """
    error = True
    if status == 'request':
      error = False
    if self.next_mtype is None and 'init' in mtype :
      error = False
    if isinstance( self.next_mtype, str ):
      if mtype == self.next_mtype:
        error = False
    elif isinstance( self.next_mtype, list ):
      if mtype in self.next_mtype:
        error = False

    if error is True:
      raise LURKError( 'invalid_type', f"unexpected request {mtype} "\
              f"expecting {self.next_mtype} or initial request" )

  def serve( self, payload, mtype, status ):
    self.is_expected_message( mtype, status )
    if mtype == 's_init_cert_verify':
      req = SInitCertVerifyReq( payload, self.conf )
      self.scheduler = req.scheduler
      self.handshake = req.handshake
      self.session_id = req.session_id
      self.last_exchange = req.tag.last_exchange
    elif mtype == 's_new_ticket':
      req = SNewTicketReq( payload, self.conf, self.handshake,\
              self.scheduler, self.session_id, self.ticket_counter )
      self.ticket_counter = req.ticket_counter
      self.last_exchange = req.tag.last_exchange
    elif mtype == 's_init_early_secret':
      req = SInitEarlySecretReq( payload, self.conf )
      self.scheduler = req.scheduler
      self.handshake = req.handshake
      self.session_id = req.session_id
      self.freshness = req.freshness
      self.session_ticket = req.session_ticket
      self.ephemeral = req.ephemeral
    elif mtype == 's_hand_and_app_secret':
      req = SHandAndAppSecretReq( payload, self.conf, self.handshake,\
              self.scheduler, self.session_id, self.session_ticket, \
              self.freshness, self.ephemeral )
      self.last_exchange = req.tag.last_exchange
    else:
      raise LURKError( 'invalid_request', f"unexpected request {mtype}"\
              f"expecting {self.next_mtype} or initial request" )
    self.next_mtype = req.next_mtype
    return req.resp



class CInitClientFinishedReq:

  def __init__(self, req, tls13_conf, debug=None ):
    """ computes the signature for the client

    If post handshake authentication is not enabled,
    than, the c_init_client_finished exchange is not
    followed by other exchange and ther eis only the
    need to perform the signature over the ClientHello
    ... later of server Finished/EndOfEarlyData.

    However, if post handshake authentication, is enabled,
    there is a need to generate the client Finished
    message, to be ready for the post authentication.
    This client Finished message requires h_c to be
    computed and so a key schedule to be initiated.
    This key schedule requires the shared secret and psk
    as input.
    """
    self.conf = tls13_conf
    self.debug = debug
    self.mtype = 'c_init_client_finished'
    self.next_mtype = 'c_post_hand_auth'


    psk = req[ 'psk' ]
    self.handshake = TlsHandshake( 'client', self.conf, debug=self.debug )
    self.handshake.msg_list.extend( req[ 'handshake' ] )
    self.handshake.sanity_check( self.mtype )
    server_cert = LurkCert( req[ 'server_certificate' ], self.mtype, self.conf, \
                    True, self.handshake )
    client_cert = LurkCert( req[ 'client_certificate' ], self.mtype, self.conf, \
                    False, self.handshake )
    self.handshake.update_random( Freshness( req[ 'freshness' ] ) )
    if server_cert.cert_type != 'no_certificate' :
      self.handshake.update_certificate( server_cert, server=True )
    self.scheduler = None
    self.tag = Tag( req[ 'tag' ], self.mtype, self.conf, ctx=self.handshake )
    self.session_id = SessionID( req[ 'session_id' ], tag=self.tag )
    ephemeral = Ephemeral( req[ 'ephemeral' ], self.mtype, self.conf, self.handshake )
    self.scheduler = KeyScheduler( self.handshake.get_tls_hash(), \
                                   shared_secret=ephemeral.shared_secret, psk=psk, debug=debug )
    self.scheduler.process( [ 'h_c', 'h_s' ], self.handshake )
    pylurk.debug.print_bin( 'h_c', self.scheduler.secrets[ 'h_c' ] )
    pylurk.debug.print_bin( 'h_s', self.scheduler.secrets[ 'h_s' ] )
    self.client_cert  = LurkCert( req[ 'client_certificate' ], \
                                  self.mtype, self.conf, server=False, \
                                  handshake=self.handshake )

    self.handshake.update_certificate( self.client_cert, server=False )
    self.handshake.update_certificate_verify( )
    ## get sig from freshly generated certificate_verify
    sig = self.handshake.msg_list[ 0 ][ 'data' ]['signature' ]
    ## generating Finished message and generating the transcript of the full handshake
    if self.tag.last_exchange is False:
##      self.handshake.update_server_finished( self.scheduler )
      self.handshake.update_finished( self.scheduler )
      self.scheduler.process( [ 'r' ], self.handshake )
    self.resp = { 'tag' : self.tag.resp,
                  'session_id' : self.session_id.cs,
                  'signature' : sig }


class CPostHandAuthReq:

  def __init__(self, req, tls13_conf, handshake, scheduler,\
               session_id, post_hand_auth_counter, debug=None ):
    self.conf = tls13_conf
    self.mtype = 'c_post_hand_auth'
    self.next_mtype = 'c_post_hand_auth'

    self.handshake = handshake
    self.handshake.msg_list.extend( req[ 'handshake' ] )
    self.handshake.update_certificate( req[ 'certificate' ], server=False )
    self.handshake.sanity_check( self.mtype )
    self.post_hand_auth_counter = post_hand_auth_counter + 1
    self.scheduler = scheduler

    ## the signature is performed for a post handshake transcript
    self.handshake.update_certificate_verify( )
    sig = self.handshake.msg_list[ 0 ][ 'data' ]['signature' ]
    del self.handshake.msg_list[ : ]

    self.tag = Tag( req[ 'tag' ], self.mtype, self.conf, ctx=self.post_hand_auth_counter )
    self.resp =  { 'tag' : self.tag.resp,
                   'session_id' : session_id.e,
                   'signature' : sig }

class CInitClientHelloReq:

  def __init__(self, req, tls13_conf, ticket_db, debug=None ):
    self.conf = tls13_conf
    self.debug = debug
    self.mtype = 'c_init_client_hello'
    self.next_mtype = [ 'c_server_hello', 'c_client_finished' ]

    self.handshake = TlsHandshake( 'client', self.conf, debug=self.debug )
    self.handshake.msg_list.extend( req[ 'handshake' ] )
    self.handshake.update_random( Freshness( req[ 'freshness' ] ) )
    ## keyshare
    self.ephemeral = Ephemeral( { 'method' : 'cs_generated' }, self.mtype, \
      self.conf, self.handshake, debug=debug )

    self.handshake.update_key_share( [ k.ks_entry() for k in self.ephemeral.client_ecdhe_key_list ])
    ## pre-shared-key extension
    secret_list = []
    binder_key_list = []
    ## the dictionaries are only filled when PSK are proposed and
    ## hosted by the CS.
    self.scheduler_list = []
    psk_metadata_list = req[ 'psk_metadata_list' ]
    self.handshake.is_post_hand_auth_proposed( )
    if self.handshake.is_psk_proposed( ) is True:
      ticket_info_list = self.handshake.psk_info_list_from_identities( \
              psk_metadata_list, ticket_db )
      self.scheduler_list = self.handshake.binder_scheduler_key_list( \
              ticket_info_list )
      binder_key_list = [ ks.secrets[ 'b' ] for ks in self.scheduler_list  ]
      binder_finished_key_list = [ ks.finished_key( role='binder' ) for ks in self.scheduler_list  ]
      self.handshake.update_binders( ticket_info_list, binder_finished_key_list )
#      secret_request = SecretReq( req[ 'secret_request' ] , self.mtype, self.conf, \
#                       handshake=self.handshake )
      secret_request = SecretReq( req[ 'secret_request' ] ,\
                                  self.mtype, self.conf )
      ## we need to provide the hash function because handshake
      ## define the hash fcuntion from the ciphersuite. The ciphersuite
      ## is in the serverHello
      ks = self.scheduler_list[ 0 ]
      self.handshake.transcript = Hash( ks.tls_hash )
      ks.process( secret_request.of( [ 'e_s', 'e_x' ] ), self.handshake )
      secret_list = secret_request.resp( ks )
      ## avoiding redundant binder_key
      if len( binder_key_list ) != 0:
        for secret in secret_list:
          if secret[ 'secret_type' ] == 'b':
            secret[ 'secret_data' ] = b''
            break

    self.session_id = SessionID( req[ 'session_id' ] )
    self.resp = { \
      'session_id' : self.session_id.cs,
      'ephemeral_list' : self.ephemeral.resp,
      'binder_key_list' : [ \
        { 'secret_type' : 'b', 'secret_data' : b } for b in binder_key_list  ],
      'secret_list' : secret_list }

### self has been forgotten
## is the function called somewhere ?
#  def update_binders( self, psk_metadata_list ):
#    ticket_info_list = self.tls_handshake.psk_info_list_from_identities( \
#            psk_metadata_list, self.ticket_db )
#    scheduler_list = self.tls_handshake.binder_sheduler_key_list( ticket_info_list )
#    binder_finished_key_list = [ks.finished_key( role='binder' ) for ks in scheduler_list ]
#    truncated_client_hello = self.tls_handshake.truncated_client_hello( ticket_info_list )
#    self.tls_handshake.update_binders( ticket_info_list, binder_finished_key_list )
#    return scheduler_list


class CServerHelloReq:

  def __init__(self, req, tls13_conf, handshake, client_hello_ephemeral,\
               scheduler_list, session_id, debug=None ):
    self.conf = tls13_conf
    self.mtype = 'c_server_hello'
    self.next_mtype = 'c_client_finished'

    self.handshake = handshake
    self.handshake.msg_list.extend( req[ 'handshake' ] )
    self.handshake.sanity_check( self.mtype )
    secret_req = { 'h_s' : True, 'h_c' : True }
    for k in [ 'b', 'e_s', 'e_x', 'a_c', 'a_s', 'x', 'r' ]:
      secret_req[ k ] = False
#    secret_request = SecretReq( secret_req, self.mtype, self.conf, \
#                                handshake=self.handshake )
    secret_request = SecretReq( secret_req, self.mtype, self.conf )
    ephemeral = Ephemeral( req[ 'ephemeral' ], self.mtype, \
                                self.conf, self.handshake )
    ephemeral.client_shared_secret( client_hello_ephemeral )
    if debug is not None:
      debug.handle_bin( 'ecdhe_shared_secret', ephemeral.shared_secret )
    if self.handshake.is_psk_agreed( ):
      selected_identity = self.handshake.server_hello_ext_data( 'pre_shared_key' )
      self.scheduler = scheduler_list[ selected_identity ]
      self.scheduler.shared_secret = ephemeral.shared_secret
    else:
      self.scheduler = KeyScheduler( self.handshake.get_tls_hash(), \
                                     shared_secret=ephemeral.shared_secret, debug=debug )
    self.scheduler.process( [ 'h_c', 'h_s' ], self.handshake )
    self.session_id = session_id
    self.resp = { 'session_id' : session_id.e,
                  'secret_list' : secret_request.resp( self.scheduler ) }


class CClientFinishedReq:

  def __init__(self, req, tls13_conf, handshake, scheduler, session_id, debug=None ):
    self.conf = tls13_conf
    self.mtype = 'c_client_finished'
    self.next_mtype = [ 'c_post_hand_auth', 'c_register_ticket' ]
    ## scheduler is None when not being selected during the c_server_hello
    ## This means that only the certificate verify (signature) is generated
    self.scheduler = scheduler
    self.handshake = handshake
    self.handshake.msg_list.extend( req[ 'handshake' ] )
    self.handshake.sanity_check( self.mtype )
    is_cert_req = self.handshake.is_certificate_request()

    self.tag = Tag( req[ 'tag' ], self.mtype, self.conf, ctx=self.handshake )
    server_cert = LurkCert( req[ 'server_certificate' ], self.mtype, self.conf, \
                    True, self.handshake )
    client_cert = LurkCert( req[ 'client_certificate' ], self.mtype, self.conf, \
                    False, self.handshake )
    if server_cert.cert_type != 'no_certificate' :
      self.handshake.update_certificate( server_cert, server=True )
    ## scheduler is initialized in CServerHello and CServerHello
    ## may be skipped when the client knows PSK and ECDHE secrets
    ## In this situation, the sheduler is not needed.
    if self.scheduler is None: ## we need to initialize transcript
      self.handshake.transcript = Hash( self.handshake.get_tls_hash() )
    else:
      self.scheduler.process( [ 'a_c', 'a_s' ], self.handshake )
    if is_cert_req:
      self.handshake.update_certificate( client_cert, server=False )
      self.handshake.update_certificate_verify( )
      ## get sig from freshly generated certificate_verify
      sig = self.handshake.msg_list[ 0 ][ 'data' ]['signature' ]
    else:
      sig = b''
    ## generating Finished message and generating the transcript of the full handshake
    if self.tag.last_exchange is False:
      self.handshake.update_finished( self.scheduler )
      self.scheduler.process( [ 'r' ], self.handshake )
    secret_request = SecretReq(req[ 'secret_request' ], self.mtype, self.conf )
    self.resp = { 'tag' : self.tag.resp,
                  'session_id' : session_id.cs,
                  'signature' : sig,
                  'secret_list' : secret_request.resp( self.scheduler ) }


class CRegisterTicketsReq:

  def __init__( self, req, tls13_conf, ticket_db, scheduler,\
                handshake, ticket_counter, session_id, debug=None ):
    self.conf = tls13_conf
    self.mtype = 'c_register_tickets'
    self.next_mtype = [ 'c_post_hand_auth', 'c_register_ticket' ]

    self.ticket_counter = ticket_counter
    self.tag = Tag( req[ 'tag' ], self.mtype, self.conf, ctx=self.ticket_counter )
    for new_session_ticket in req[ 'ticket_list' ]:
      self.ticket_counter += 1
      if self.ticket_counter > self.conf[ 'max_tickets' ]:
        raise LURKError( 'max_tickets_reached' )
      ticket_db.register( new_session_ticket, scheduler, handshake )


    self.resp = { 'tag' : self.tag.resp,
                  'session_id' : session_id.cs }

class CSession(SSession) :

  def __init__( self, tls13_conf:dict, session_db=None,\
                ticket_db=None, debug=None ):
    super().__init__( tls13_conf=tls13_conf, session_db=session_db, \
                      ticket_db=ticket_db, debug=debug )
    self.post_hand_auth_counter = None
    self.scheduler_list = None
    self.transcript_r = None
    self.cipher_suite = None
    self.tls_hash = None

  def serve( self, payload, mtype, status ):
    ## check request and next_
    self.is_expected_message( mtype, status )
    if mtype == 'c_init_client_finished':
      req = CInitClientFinishedReq( payload, self.conf, debug=self.debug )
      self.scheduler = req.scheduler
      self.handshake = req.handshake
      self.session_id = req.session_id
      self.last_exchange = req.tag.last_exchange
      self.post_hand_auth_counter = 0
    elif mtype == 'c_post_hand_auth':
      req = CPostHandAuthReq( payload, self.conf, self.handshake, self.scheduler,\
                           self.session_id, self.post_hand_auth_counter, debug=self.debug )
      self.last_exchange = req.tag.last_exchange
      self.post_hand_auth_counter = req.post_hand_auth_counter
    elif mtype == 'c_init_client_hello':
      req = CInitClientHelloReq( payload, self.conf, self.ticket_db, debug=self.debug )
      self.scheduler_list = req.scheduler_list
      self.handshake = req.handshake
      self.session_id = req.session_id
      self.ephemeral = req.ephemeral
      self.scheduler_list = req.scheduler_list
    elif mtype == 'c_server_hello' :
      req = CServerHelloReq( payload, self.conf, self.handshake, self.ephemeral,\
                             self.scheduler_list, self.session_id, debug=self.debug )
      ## maybe we can remove these two lines
      self.handshake = req.handshake
      self.scheduler = req.scheduler
    elif mtype == 'c_client_finished' :
      req = CClientFinishedReq( payload, self.conf, self.handshake, self.scheduler,\
                             self.session_id, debug=self.debug )
      ## saving context for the register_tickets
      self.transcript_r = req.handshake.transcript_r
      self.cipher_suite = req.handshake.cipher_suite
      ## the scheduler is set to None when the TLS client knwos both PSK and
      ## ECDHE shared secret. In that case, last_exchange MUST be set to True
      if req.scheduler is not None:
        self.tls_hash = req.scheduler.tls_hash
      self.last_exchange = req.tag.last_exchange
      self.ticket_counter = 0
    elif mtype == 'c_register_tickets' :
      req = CRegisterTicketsReq( payload, self.conf, self.ticket_db,\
              self.scheduler, self.handshake,\
              self.ticket_counter, self.session_id, debug=self.debug )
      self.ticket_counter = req.ticket_counter
    else:
      raise LURKError( 'invalid_request', "unexpected request {mtype}" )
    self.next_mtype = req.next_mtype
    return req.resp

class SessionDB:

  def __init__(self, ):

    self.db = {}

  def store( self, session:SSession):
    self.db[ session.session_id.cs ] = session

  def unstore( self, session_id:bytes ):
    return self.db[ session_id ]

  def delete( self, session:SSession):
    del self.db[ session.session_id.cs ]

class TicketDB:

  def __init__( self, debug=None ):
    self.db = {}
    self.debug=debug

  def key( self, new_session_ticket ):
    return new_session_ticket[ 'ticket' ]

  def register( self, new_session_ticket, ks, tls_handshake ):
    key = self.key( new_session_ticket )
    self.db[ key ] = { 'new_session_ticket' : new_session_ticket,
                    'psk_bytes' : ks.compute_psk( new_session_ticket[ 'ticket_nonce' ] ),
                    'tls_hash' : tls_handshake.get_tls_hash(),
                    'psk_type' : 'resumption',
                    'cipher_suite' : tls_handshake.cipher_suite,
                    'registration_time' : time.time() }
    if self.debug is not None:
      self.debug.trace_val( "Registering in CS Tickert DB", self.db[ key ] )

  def update_ticket_info( self, ticket_info ):
    """ update ticket_info with obfuscated_ticket_age

    This function updates the information so all information is
    available to generate a TLS ticket
    """
    if ticket_info[ 'psk_type' ] == 'external':
      obfuscated_ticket_age = 0
    else :
      obfuscated_ticket_age = int( ( time.time( ) - \
              ticket_info[ 'registration_time' ] ) * 1000 +\
              ticket_info[ 'new_session_ticket' ][ 'ticket_age_add' ] ) % 2**32
    ticket_info[ 'obfuscated_ticket_age' ] = obfuscated_ticket_age
    del ticket_info[ 'registration_time' ]
    return  ticket_info

  def get_ticket_info( self, ticket ):
    """ returns the ticket_info associated to ticket

    The intent of the function is to correlate the structure of
    the ticket DB and the key being used to retrieve the ticket_info.
    """
    try:
      return self.update_ticket_info( self.db[ ticket ] )
    except KeyError:
      return None

class Tls13Ext:
  def __init__(self, conf, ticket_db=None, session_db=None, debug=None ):
    self.conf = conf
    self.debug = debug
    if session_db is None:
      self.session_db = SessionDB()
    else:
      self.session_db = session_db
    if ticket_db is None:
      self.ticket_db = TicketDB( debug=self.debug)
    else:
      self.ticket_db = ticket_db

  def payload_resp( self, req:dict ) -> dict :
    req_type = req[ 'type' ]
    req_payload = req[ 'payload' ]
    if 'init' in req_type : #in [ 's_init_early_secret', 's_init_cert_verify' ]:
      if req_type[ :7 ] == 's_init_':
        session = SSession( self.conf, session_db=self.session_db,\
                            ticket_db=self.ticket_db, \
                            debug=self.debug )
      elif req_type[ :7 ] == 'c_init_' :
        session = CSession( self.conf, session_db=self.session_db,\
                            ticket_db=self.ticket_db,\
                            debug=self.debug )
      else:
        raise LURKError( 'invalid_type', f"{req_type}" )
      payload  = session.serve( req_payload, req_type, 'request')
      if session.session_id is not None :
        self.session_db.store( session )
    else :
      try:
        session = self.session_db.unstore( req_payload[ 'session_id' ] )
      except KeyError:
        raise LURKError( 'invalid_session_id', f"{req} session_id not found in DB" )
      payload = session.serve( req_payload, req_type, 'request' )
    if session.last_exchange is True:
      self.session_db.delete( session )
    return payload




import os.path
import binascii
import json
import pprint
from pylurk.struct_lurk import LURKMessage
from pylurk.lurk.lurk_lurk import ConfigurationError

def get_struct(struct_list: list, key:str, value) -> dict :
  """ return the first element of the list that contains  key:value """
  for e in struct_list:
    if e[ key ] == value:
      return e
  return None

def get_struct_index(struct_list: list, key:str, value) -> dict :
  """ return the first element of the list that contains  key:value """
  for e in struct_list:
    if e[ key ] == value:
      return struct_list.index(e)
  return None

def str_to_bytes( hexlify_string:str ):
  """ converts string to bytes 

  This could be useful for example when bytes needs to be 
  stored into a json file. JSON does not store bytes, so 
  these needs to be stored as bytes.  

  The reverse function is bytes_to_str.

  Args:
    hexlify_string: the string that represents bytes - 
      it has a specific format

  Returns:
    bytes_output: the represented bytes (by the string).
  """
  hexlify_string = str( hexlify_string )
  bytes_output = b''
  for hex_str in hexlify_string.split( " " ):
    bytes_output += binascii.unhexlify( hex_str )
  return bytes_output

def bytes_to_str( bit_string ):
  """ converts bytes to string 

  Args:
    bit_string: th ebytes stream

  Returns:
    the string representaing the bytes
  """
  return binascii.hexlify(bit_string, sep=' ').decode()

def bytes_to_human( description:str, bit_string:bytes ):
  """ display bytes stream into a human readable format

  Args: 
    description: a string that describes the coming bytes
    bit_string: the byte stream to display

  Returns:
    a string
  """

  output_string = f"  - {description} [{len(bit_string)} bytes]:\n"
  sep=0
  for char in bytes_to_str( bit_string ):
    if char == ' ':
      sep += 1
      if sep % 16 == 0:
        output_string += '\n'
      else:
        output_string += char
    else:
      output_string += char
  return output_string

def check_file( file_path ):
  """ checks a file can be open """
  if file_path is None :
    raise ConfigurationError( "file_path is None. Expecting file path." )
  if os.path.exists( file_path ) is False:
    raise ConfigurationError( f"{file_path} does not exist." )
  if os.path.isfile( file_path ) is False:
    raise ConfigurationError( f"{file_path} is not a file but a directory" )
  if os.stat( file_path ).st_size == 0 :
    raise ConfigurationError( f"{file_path} is empty" )

def print_bin( description:str, bit_string:bytes ):
  """ prints human readable bytes to the stdout 

  Args: 
    description: a string that describes the coming bytes
    bit_string: the byte stream to display
  """

  print( bytes_to_human( description, bit_string ) )

def print_val( key:str, value ):
  """ pretty print values

  We should probably replace print_bin by print val
  """

  if isinstance( value , ( bytes, bytearray ) ):
    ##    print( f"  - {key} [{len(value)} bytes]: {value}" )
    print_bin( key, value )
  else:
    pprint.pprint( f"  - {key}: {value}", width=80, sort_dicts=False )

class Debug:

  def __init__( self, debug_conf ):
    self.conf = debug_conf
    self.test_vector = False
    self.test_vector_file = None
    test_vector_mode = None
    self.check = False
    self.record = False
    self.db = {}
    self.trace = False
    if 'trace' in debug_conf.keys( ):
      self.trace = debug_conf[ 'trace' ]
      if isinstance( self.trace, bool ) is False:
        raise ConfigurationError( f"Unexpected value for trace."\
          f" Expecting boolean value. {debug_conf}" )

    ## handling test_vector when present test_vector MUST have a file
    ## and a mode.
    if 'test_vector' in debug_conf.keys( ):
      key_list = debug_conf[ 'test_vector' ]
      if 'file' in key_list and 'mode' in key_list :
        self.test_vector_file = debug_conf[ 'test_vector' ][ 'file' ]
        test_vector_mode = debug_conf[ 'test_vector' ][ 'mode' ]
        self.test_vector = True
        if test_vector_mode == 'check':
          self.check = True
        elif test_vector_mode == 'record':
          self.record = True
        elif test_vector_mode is None :
          pass
        else:
          raise ConfigurationError( f"Unexpected mode. Acceptable values "\
            f"are: 'check', 'record' or None. Provided mode is {debug_conf}" )

        if os.path.exists( self.test_vector_file ) and\
           os.path.isfile( self.test_vector_file ) and\
           os.stat( self.test_vector_file ).st_size > 0:
          with open( self.test_vector_file, 'rt', encoding='utf8' ) as f:
            self.db = json.load( f )

  def read_bin( self, key ) -> bytes :
    """ returns the value in a binary format """
    if key in self.db.keys() :
      return str_to_bytes( self.db[ key ] )
    return None

  def read_str( self, key ) -> str :
    """ returns the value in a string representation"""
    if key in self.db.keys() :
      return self.db[ key ]
    return None

  def check_bin( self, key:str, value:bytes ):
    """ raises an error when the key, value mismatches those of the test_vector """
    if key in self.db.keys() :
      ref_value = str_to_bytes( self.db[ key ] )
      if ref_value != b'' :
        if value != ref_value :
          raise ValueError(
            f"Debug {key} check fails:\n"\
            f"{bytes_to_human( 'expected', ref_value)}\n"\
            f"{bytes_to_human( 'provided', value)}\n" )

  def value_to_json( self, struct:dict ):
    """ stores a given value into a compatible json format

    JSON files are very restrictive regarding the type of data
    they can store. More specifically, they cannot store bytes,
    so we have to convert these bytes into strings. 
    This fonction is able to convert any structure into a JSON
    compatible structure, that is to say, it is able to look 
    for the bytes stream into that structure.

    Args:
      struct: a dictionnary or list that can conatin a bytesstream. 
        In general dictionnary are expected to represent complex
        structures.

    Returns:
      a JSON compatible structure, that can be stored into a JSON file.
    """
    if isinstance( struct, dict ) is True:
      for k in struct.keys():
        v = struct[ k ]
        if isinstance( v, ( bytes, bytearray ) ):
          struct[ k ] = bytes_to_str( v )
        elif isinstance( v, dict ):
          struct[ k ] = self.value_to_json( v )
    elif isinstance( struct, list ) is True:
      for i in range( len ( struct ) ):
        struct[ i ] = self.value_to_json( struct[ i ] )
    return struct

  def record_bin( self, key:str, value:bytes ):
    value = bytes_to_str( value )
    self.record_val( key, value )

  def record_val( self, key:str, value ):
    """ record key value to the test_vector file

    The value is simultaneously added to the db as well as to the
    test_vector file.
    The reason to do so is to allow a common test_vector file being
    shared by the multiple modules as well as to avoid synchronizing
    the various db.
    """
    ## updating the local db
    self.db[ key ] = self.value_to_json( value )

    ## updating the test_vector file
    ## opens only when the file exists to read the existing records.
    if os.path.exists( self.test_vector_file ) and os.path.isfile( self.test_vector_file )\
       and os.stat( self.test_vector_file ).st_size > 0:
      with open( self.test_vector_file, 'rt', encoding='utf8' ) as f:
        tmp_db = json.load( f )
#        tmp_db[  key ] = self.db[ key ]
    else:
      tmp_db = {}
    tmp_db[ key ] = self.value_to_json( value )
    with open( self.test_vector_file, 'wt', encoding='utf8' ) as f:
      json.dump( tmp_db, f, indent=2 )

  def trace_bin( self, key:str, value:bytes ):
    print( bytes_to_human( key, value ) )

  def trace_val( self, key:str, value:bytes ):
    if isinstance( value , ( bytes, bytearray ) ):
      print( f"  - {key} [{len(value)} bytes]: {value}" )
    else:
#      pprint.pprint( f"  - {key}: {value}", width=80, sort_dicts=False )
      print( f"  - {key}: {value}" )

  def handle_bin( self, key:str, value:bytes ):
    if self.check is True:
      self.check_bin( key, value )
    if self.record is True:
      self.record_bin( key, value )
    if self.trace is True:
      self.trace_bin( key, value )


class Tls13Debug( Debug ):


  def check_client_ephemeral( self, ephemeral ):
    for k in ephemeral.client_ecdhe_key_list:
      if k.private is not None:
        self.check_bin( f"client_{k.group}_ecdhe_private", k.pkcs8() )

  def trace_client_ephemeral( self, ephemeral ):
    for k in ephemeral.client_ecdhe_key_list:
      if k.private is not None:
        pkcs8_bytes = k.pkcs8( )
        key = f"client_{k.group}_ecdhe_private"
        self.trace_bin( key, pkcs8_bytes )
        self.trace_val( key, pkcs8_bytes )
      self.trace_val( f"client_{k.group}_ecdhe_public", k.ks_entry( ) )

  def record_client_ephemeral( self, ephemeral ):
    for k in ephemeral.client_ecdhe_key_list:
      if k.private is not None:
        self.record_bin( f"client_{k.group}_ecdhe_private", k.pkcs8() )

  def handle_client_ephemeral( self, ephemeral ):
    if self.check is True:
      self.check_client_ephemeral( ephemeral )
    if self.record is True:
      self.record_client_ephemeral( ephemeral )
    if self.trace is True:
      self.trace_client_ephemeral( ephemeral )

  def lurk_msg_key( self, struct:dict ):
    return f"{struct[ 'type' ]}_{struct[ 'status' ]}"

  def check_lurk_msg( self, msg:dict ):
    pass
###    msg_bytes = LURKMessage.build( msg )
###    self.check_bin( f"{self.lurk_msg_key( msg )}_bytes", msg_bytes )

  def trace_lurk_msg( self, msg:dict ):
    key = self.lurk_msg_key( msg )
    msg_bytes = LURKMessage.build( msg )
    self.trace_val( f"{key}", LURKMessage.parse( msg_bytes ) )
    self.trace_bin( f"{key}_bytes", msg_bytes )

  def record_lurk_msg( self, msg ):
    key = self.lurk_msg_key( msg )
    msg_bytes = LURKMessage.build( msg )
    self.record_bin( f"{key}_bytes", msg_bytes )
##    self.record_val( f"{key}", dict( msg ) )

  def handle_lurk_msg( self, msg ):
    if self.check is True:
      self.check_lurk_msg( msg )
    if self.record is True:
      self.record_lurk_msg( msg )
    if self.trace is True:
      self.trace_lurk_msg( msg )




from construct.core import *
from construct.lib import *

LURKVersion = Enum( BytesInteger( 1 ),
  v1 = 1
)

## LURK Type
LURKType = Enum( BytesInteger( 1 ) ,
  capabilities = 0,
  ping = 1,
  error = 2
)

## LURK status
LURKStatus = Enum( BytesInteger( 1 ),
  request = 0,
  success = 1,
  undefined_error = 2,
  invalid_format = 3,
  invalid_extension = 4,
  invalid_type = 5,
  invalid_status = 6,
  temporary_failure = 7
)

############  LURKErrorResponse

ErrorPayload = Struct (
  'lurk_state' / Bytes(4), \
)

########### Empty Payload
EmptyPayload = Struct()



#from secrets import randbits, token_bytes
#from  os.path import join
#from copy import deepcopy
#import datetime
#import hashlib



from typing import  NoReturn
#Union, TypeVar, List

from cryptography.hazmat.primitives.asymmetric import utils
#from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
#from cryptography.hazmat.primitives.asymmetric.x448 import X448PrivateKey, X448PublicKey

#from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey
#from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PrivateKey, Ed448PublicKey
#from cryptography.hazmat.primitives.asymmetric import rsa
#from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
#from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.ec import ECDSA
#SECP256R1, SECP384R1, SECP521R1,
#EllipticCurvePublicNumbers, ECDH, EllipticCurvePrivateKey, EllipticCurvePublicKey

from cryptography.hazmat.primitives.hashes import Hash
#, SHA256, SHA384, SHA512
#from cryptography.hazmat.primitives.serialization import load_der_private_key,
#load_pem_private_key, NoEncryption, Encoding, PrivateFormat, PublicFormat
#from cryptography import x509
#from cryptography.x509.oid import NameOID

#from cryptography.hazmat.primitives.asymmetric import padding
#from cryptography.hazmat.primitives import serialization

from cryptography.hazmat.primitives.hmac import HMAC

#from construct.core import *
#from construct.lib import *


import pytls13.struct_tls13 as tls
import pylurk.debug
from pylurk.debug import get_struct
#, get_struct_index
from pylurk.tls13.crypto_suites import SigScheme, CipherSuite
from pylurk.lurk.lurk_lurk import LURKError, ImplementationError
import pylurk.tls13.key_scheduler


class TlsHandshake:

  def __init__( self, role, tls13_conf=None, debug=None ) -> None:
    self.role = role #tls13_conf[ 'role' ]  ## list of role
    ## mostly makes sense for the server side
    self.conf = tls13_conf
    if self.conf is not None :
      self.finger_print_dict = self.conf[ '_finger_print_dict' ]
      self.private_key = self.conf[ '_private_key' ]
    self.debug = debug
    ## list of structures representing the TLS handshake messages
    self.msg_list = []
    self.cipher_suite = None
    self.transcript = None
    ## define if a client certificate is requested / expected
    self.is_certificate_request_state = None
    ## define if a client has claimed supporting post handshake authentication
    self.post_hand_auth_proposed = None
    ## transcript of the handshake
    ## to be re-used for the resumption secret and post handshake authentication
    ## transcript is expresssed in bytes.
    self.transcript_r = None
    self.ks_proposed = None
    self.psk_proposed = None

  def msg_type_list( self ):
    """ returns the list of message types

    This is usefull to check a messgage is present in the handshake
    """
    return [ msg[ 'msg_type' ] for msg in self.msg_list ]

  def latest_client_hello_index( self ):
    """ return the index of the latest client hello

    The following cases are considered:
      * [ clienthello]
      * [ clientHello, hello_retry_request]
      * [ clientHello, hello_retry_request, client_hello ]
      * [ client hello, server]
    """
    if self.msg_list[ 0 ][ 'msg_type' ] != 'client_hello' :
      raise LURKError( 'invalid_handshake', f"Expecting client hello"\
                       f" in first message {self.msg_list}" )
    if len( self.msg_list ) <= 2 :
      index = 0
    elif len( self.msg_list ) >= 3 :
      if self.msg_list[ 2 ][ 'msg_type' ] == 'client_hello':
        index = 2
      else:
        index = 0
    return index

  def server_hello_index( self ):
    """ returns the index of the server_hello message """
    ## when self,msg_list starts with serverhello: s_hand_and_app
    if self.msg_list[ 0 ][ 'msg_type' ] == 'server_hello' :
      sh_index = 0
    # when self.msg_list starts with client hello: s_init_cert_verify,
    else:
      ch_index = self.latest_client_hello_index()
      if len( self.msg_list ) < ch_index + 1:
        raise ImplementationError( f"cannot find server hello {self.msg_list}" )
      sh_index = ch_index + 1
    return sh_index

  def client_hello_extension_list( self ):
    ch_index = self.latest_client_hello_index()
    ext_list = []
    for ext in self.msg_list[ ch_index ][ 'data' ][ 'extensions' ] :
      ext_list.append( ext[ 'extension_type' ] )
    return ext_list

  def server_hello_extension_list( self ):
    sh_index = self.server_hello_index()
    ext_list = []
    for ext in self.msg_list[ sh_index ][ 'data' ][ 'extensions' ] :
      ext_list.append( ext[ 'extension_type' ] )
    return ext_list

  def client_hello_ext_data( self, extension_type ):
    ch_index = self.latest_client_hello_index()
    for ext in self.msg_list[ ch_index ][ 'data' ][ 'extensions' ] :
      if ext[ 'extension_type' ] == extension_type:
        return ext[ 'extension_data' ]
    return None

  def server_hello_ext_data( self, extension_type):
    ch_index = self.server_hello_index()
    for ext in self.msg_list[ ch_index ][ 'data' ][ 'extensions' ] :
      if ext[ 'extension_type' ] == extension_type:
        return ext[ 'extension_data' ]
    return None


  def is_psk_proposed( self )->bool :
    """ return True if self.msg_list has proposed PSK, False otherwise """
    if self.psk_proposed is not None:
      return self.psk_proposed

    ext_list = self.client_hello_extension_list( )
    if 'pre_shared_key' in ext_list  and  'psk_key_exchange_modes' in ext_list :
      self.psk_proposed = True
    else:
      self.psk_proposed = False
    return self.psk_proposed

  def is_psk_agreed( self ) -> bool :
    """ return True is PSK has been agreed, False otherwise """
    ext_list = self.server_hello_extension_list( )
    if 'pre_shared_key' in ext_list :
      psk_agree = True
    else:
      psk_agree = False
    return psk_agree

  def is_ks_proposed( self )->bool :
    """ return True if a key share extension is in the client_hello """

    if self.ks_proposed is not None:
      return self.ks_proposed

    ext_list = self.client_hello_extension_list( )
    if 'key_share' in ext_list :
      self.ks_proposed = True
    else:
      self.ks_proposed = False
    return self.ks_proposed

  def is_ks_agreed( self ) -> bool :
    """ return True if a key_share extension is in the server hello """
    if 'key_share' in self.server_hello_extension_list( )  :
      return True
    return False

  def is_certificate_request( self ):
    """ returns True is the handshake contains a CertificateRequest message

    The checks performed by the TLS client or the TLS server are a bit different
    as the tLS client also checks the presence of the CertificateVerify.
    """
    if self.is_certificate_request_state is None:
      if 'certificate_request' in self.msg_type_list( ):
        self.is_certificate_request_state = True
      else:
        self.is_certificate_request_state = False
    return self.is_certificate_request_state

  def is_certificate_agreed( self ):
    """ return True if the handshake is acceptable for a certificate authentication  """
    if self.is_ks_agreed( ) is True :
      if 'signature_algorithms' in self.client_hello_extension_list( ):
        return True
    return False

  def is_early_data_proposed( self ):
    """ returns True is the TLS client enable early_data """
    ch_index = self.latest_client_hello_index()
    ext_list = []
    for ext in self.msg_list[ ch_index ][ 'data' ][ 'extensions' ] :
      ext_list.append( ext[ 'extension_type' ] )
    if 'early_data' in ext_list :
      return True
    return False

  def is_post_hand_auth_proposed( self ):
    """ returns True if the TLS client supports post handshake authentication """
    if self.post_hand_auth_proposed is  None:
      if 'post_handshake_auth' in self.client_hello_extension_list( ):
        self.post_hand_auth_proposed = True
      else:
        self.post_hand_auth_proposed = False
    return self.post_hand_auth_proposed

  def sanity_check( self, mtype, session_ticket=None ):
    """ checks if the handshake is compatible with the lurk exchange """
#    error_txt = ""
    if mtype == 's_init_cert_verify':
      ## determine is a client Certificate
      ## is expected in the s_new_ticket
      self.is_certificate_request( )
      if self.is_psk_agreed() is True or self.is_ks_proposed is False or\
         self.is_ks_agreed is False:
        raise LURKError('invalid_handshake', \
                f"expecting ks_agreed and non psk_aghreed {self.msg_list}" )
## We foudn erros in the coditions checked with pylint but I cannot see how to
## updat ethe code.
#    elif mtype == 's_new_ticket' :
#      if self.is_certificate_agreed !=  self.is_certificate_request( ) is False :
#        raise LURKError('invalid_handshake', f"incompatible client "\
#                        f"Certificate / CertificateRequest" )
#      if self.resumption_master_secret is None and self.msg_list == []:
#        raise LURKError('invalid_handshake', f"expecting non empty handshake" )
    elif mtype == 's_init_early_secret':
      if self.is_psk_proposed() is False :
        raise LURKError('invalid_handshake', f"expecting psk_proposed {self.msg_list}" )
    elif mtype == 's_hand_and_app_secret':
      if self.is_psk_agreed() is False :
        raise LURKError('invalid_handshake', f"expecting psk_agreed {self.msg_list}" )
      ## checking the selected ticket indicated in the server hello is
      ## the one selected by the engine (LURK client) in the previous exchange.
      ## In the s_init_early_secrets, the engine indicates the ticket to be
      ## considered.
      ## That ticket has been used to initialize the handshake with the
      ## cipher and the tls_hash.
      ## Here we are checking the server hello actually reflect these choices.
      ## we also check the tickets ciphersuite match the one of te server hello.
      ## Note that the client hello is not available at that stage.
      server_hello_exts = self.msg_list[ 0 ][ 'data' ][ 'extensions' ]
      sh_selected_identity = get_struct( server_hello_exts, 'extension_type',\
                             'pre_shared_key' )[ 'extension_data' ]
      if sh_selected_identity != session_ticket.selected_identity :
        raise LURKError( 'invalid_handshake', f"server hello not selecting "\
                f" currenlty used ticket: server hello {sh_selected_identity} "\
                f" expecting value: {session_ticket.selected_identity}" )
      self.cipher_suite = self.msg_list[ 0 ][ 'data' ][ 'cipher_suite' ]

      if type( CipherSuite( self.cipher_suite ).get_hash() ) != type( session_ticket.tls_hash ) :
        raise LURKError( 'invalid_handshake', f"TLS handshake cipher "\
                f"suite {self.cipher_suite} and ticket cipher suites "\
                f"are not compatible {session_ticket.cipher}" )
    elif mtype == 'c_init_client_finished':
      if self.is_certificate_request is False and self.is_post_hand_auth_proposed() is False :
        raise LURKError( 'invalid_handshake', "Expecting server certificate authentication" )
    elif mtype in [ 'c_init_post_hand_auth', 'c_post_hand_auth' ]:
      if self.is_post_hand_auth_proposed() is False:
        raise LURKError( 'invalid_handshake', "Post handshake authentication no enabled" )
    elif mtype == 'c_init_client_hello':
      if self.is_psk_proposed() is False and self.is_ks_proposed() is False:
        raise LURKError( 'invalid_handshake', "psk_proposed or ks_proposed expected." )
    elif mtype == 'c_server_hello':
      pass
    elif mtype == 'c_client_finished':
      pass
    else:
      raise ImplementationError( "unknown mtype {mtype}" )


  def update_random(self, freshness):
    """ reads the engine nonce and computes the server_hello.random or client_hello.random

    """
    if self.role == 'server' : ## updating server_hello
      if self.msg_list[0][ 'msg_type' ] == 'client_hello':
        ch_index = self.latest_client_hello_index( )
        sh_index = ch_index + 1
      elif  self.msg_list[0][ 'msg_type' ] == 'server_hello' :
        sh_index = 0
      else:
        raise ImplementationError( f"Expecting handshake starting with" \
                f"client_hello or server_hello {self.msg_list}" )
      engine_random = self.msg_list[ sh_index ] [ 'data' ][ 'random' ]
      server_random = freshness.update_random( self.role, engine_random )
      self.msg_list[ sh_index ] [ 'data' ][ 'random' ] = server_random
    elif self.role == 'client':
      ch_index = self.latest_client_hello_index( )
      ch_index_list = [ 0 ]
      if ch_index != 0: ## presence of an hello retry
        ch_index_list.append( ch_index )
      engine_random = self.msg_list[ 0 ] [ 'data' ][ 'random' ]
      client_random = freshness.update_random( self.role, engine_random )
      for ch_index in ch_index_list :
        self.msg_list[ ch_index ] [ 'data' ][ 'random' ] = client_random
    else:
      raise ImplementationError( f"unknown role {self.role}" )

  def key_share_index( self, side ):
    """ return necessary arguments to point to the key_share_index

    pointers are returned (instead of simply the value) as we
    also want to update the key_share value.
    """
    if side == 'server':
      # update occurs in the server_hello
      if self.msg_list[ 0 ][ 'msg_type' ] == 'client_hello' :
        ch_index = self.latest_client_hello_index( )
        msg_index = ch_index + 1
      elif self.msg_list[ 0 ][ 'msg_type' ] == 'server_hello' :
        msg_index = 0
      else:
        raise ImplementationError( f"unexpected message type. Expecting "\
                f"client_hello or server_hello as first message "\
                f"{self.msg_list[ 0 ] [ 'msg_type' ]}" )
      ks_designation = 'server_share'
    elif side == 'client':
      # update the 'client_hello'
      msg_index = self.latest_client_hello_index( )
      ks_designation = 'client_shares'
    else:
      raise ImplementationError( f"unknown role {self.role}" )

    exts = self.msg_list[ msg_index ][ 'data' ][ 'extensions' ]
    if 'key_share' in [ e[ 'extension_type' ] for e in exts ] :
      key_share = get_struct( exts, 'extension_type', 'key_share' )
      key_share_index = exts.index( key_share )
    else:
      key_share_index = None

    return ks_designation, msg_index, key_share_index

  def get_key_share( self, side ):
    ks_designation, msg_index, ks_index = self.key_share_index( side )
    return self.msg_list[ msg_index ][ 'data' ]\
        [ 'extensions' ][ ks_index ][ 'extension_data' ]\
        [ ks_designation ]


  def update_key_share( self, key_share_entry ) -> NoReturn:
    """ update key_exchange value of the CLientHello or ServerHello

    On the server side, the key_share is the server key_share entry.
    On the client side, the key_share is a list of key_share entries (client_shares)
    """

    ks_designation, msg_index, ks_index = self.key_share_index( self.role )
    if msg_index is not None and ks_index is not None:
      self.msg_list[ msg_index ][ 'data' ]\
        [ 'extensions' ][ ks_index ][ 'extension_data' ]\
        [ ks_designation ] = key_share_entry

  def psk_info_list_from_identities( self, psk_metadata_list, ticket_db ):
    """ builds psk_info associated to each identities

    identities are always provided, psk_information are either
    retrieved from the PSKIdentityMetadata being provided or retrived from the TicketDB.

    binders are never provided to the server. This function is only to be used by the TLS server.
    """
    offered_psk = self.client_hello_ext_data( 'pre_shared_key' )
    ticket_info_list = []
    for psk_identity in offered_psk[ 'identities' ] :
      ticket_info = ticket_db.get_ticket_info( psk_identity[ 'identity' ] )
      if ticket_info is None:
        psk_identity_index = offered_psk[ 'identities' ].index( psk_identity )
        try:
          psk_metadata = psk_metadata_list.pop( 0 )
          while psk_metadata[ 'identity_index' ] < psk_identity_index:
            psk_metadata = psk_metadata_list.pop( 0 )
          if psk_metadata[ 'identity_index' ] == psk_identity_index:
            ticket_info = psk_metadata
            del ticket_info[ 'identity_index' ]
        except IndexError:
          raise LURKError( 'invalid_psk', f"Cannot retrive psk_infop for {psk_identity}" )
      ticket_info_list.append( ticket_info  )
    return ticket_info_list


  def truncated_client_hello( self, ticket_info_list ) -> bytes :
    """ returns  the necessary handshake context to generate the binders.

    The handshake context is returned in bytes which corresponds to
    Truncate(ClientHello1) or ClientHello1, HelloRetryRequest,
    Truncate(ClientHello2)

    This basically consists in removing the binders of the latest ClientHello
    """
    ## binder list is just generated to be able to built the ClientHello
    ## binders are set to zero bytes and will be removed during the truncation
    binder_list = []
    binder_list_len = 0
    for ticket_info in ticket_info_list:
      binder_len = pylurk.tls13.crypto_suites.hash_sanity_check( \
                     ticket_info[ 'tls_hash' ] ).digest_size
      binder_list.append( { 'binder' : b"\x00" * binder_len } )
      binder_list_len += 1 + binder_len ## considering binder + its size
    ## adding the binders to the latest extension which is supposed to be pre_shared_key
    ch_index = self.latest_client_hello_index( )
    self.msg_list[ ch_index ][ 'data' ][ 'extensions' ][ -1 ]\
                 [  'extension_data'] [ 'binders' ] = binder_list
    truncated_client_hello_bytes = b''
    for msg in self.msg_list[ : ch_index + 1]:
      truncated_client_hello_bytes += tls.Handshake.build( msg )
    truncated_client_hello_bytes = \
      truncated_client_hello_bytes[ : -binder_list_len - 2 ] ## removing list length
    if self.debug is not None:
      self.debug.handle_bin( 'truncated_client_hello', truncated_client_hello_bytes )
    return truncated_client_hello_bytes

  def compute_binder( self, tls_hash, binder_finished_key, \
                      truncated_client_hello_bytes, binder_index='' ) -> bytes:
    """ compute the binder associated to the binder_key

    The computation is performed similalrly to the finished message except
    that the TLS context only consists in the truncated ClientHello.

    We could have re-used the get_verify_data function that is used to
    generate the finished message. However, currently, we prefer to have
    the running hash of the handshake being decorelated from the one used
    to generate the binders.
    This function may appear as a bit redundant to get_verify_data.

    args:
      - tls_hash : the hash function used to generate the HMAC
      - binder_finished_key: the key (secret) used to generate the HMAC
      - truncated_client_hello
    """
    tls_hash = pylurk.tls13.crypto_suites.hash_sanity_check( tls_hash )
    transcript = Hash( tls_hash )
    transcript.update( truncated_client_hello_bytes )
    transcript_hash = transcript.finalize( )

    hmac = HMAC( binder_finished_key, tls_hash )
    hmac.update( transcript_hash )
    binder = hmac.finalize()
    if self.debug is not None:
      if binder_index != '':
        binder_index = f"({binder_index})"
      self.debug.handle_bin( f"compute_binder: {binder_index} "\
                             f"binder_finished_key" , binder_finished_key )
      self.debug.handle_bin( f"Transcript( truncated_client_hello ) "\
                             f"{binder_index}" , transcript_hash )
      self.debug.handle_bin( f"binder {binder_index}" , binder )

    return binder


  def binder_scheduler_key_list( self, ticket_info_list ):
    """ generates key_scheduler and binder_keys associated to tickets """
    ks_list = []
    for ticket_info in ticket_info_list :
      tls_hash = ticket_info[ 'tls_hash' ]
      psk = ticket_info[ 'psk_bytes' ]
      if ticket_info[ 'psk_type' ] == 'resumption':
        is_ext = False
      else:
        is_ext = True
      ks = pylurk.tls13.key_scheduler.KeyScheduler( tls_hash, psk=psk,\
             is_ext=is_ext, debug=self.debug )
      ks.process( [ 'b' ] , None )
      ks_list.append( ks )
    if self.debug is not None:
      for ticket_info in ticket_info_list :
        index = ticket_info_list.index( ticket_info )
        self.debug.trace_val( f"ticket ({index})", ticket_info )
        ks = ks_list[ index ]
        self.debug.handle_bin( f"binder_key ({index})", ks.secrets[ 'b' ] )
        self.debug.handle_bin( f"binder_finished_key ({index})", ks.finished_key( role='binder') )
    return ks_list

  def update_binders( self, ticket_info_list, binder_finished_key_list )-> None :
    """ generates binders to the ClientHello """
    offered_psks = self.client_hello_ext_data( 'pre_shared_key' )
    if len( offered_psks[ 'identities' ] ) != len( ticket_info_list ) or\
       len( offered_psks[ 'identities' ] ) != len( binder_finished_key_list ) :
      raise LURKError( 'invalid_psk', f"Mistmatching list lenggth: \n"\
        f"   - identitities (len: {len( offered_psks[ 'identities' ] )})\n"\
        f"   - psk_info_list  (len: {len( ticket_info_list )})\n"\
        f"   - binder_key_list  (len: {len( binder_finished_key_list )})" )

    ## generates the truncated client hello
    truncated_client_hello = self.truncated_client_hello( ticket_info_list )
    ## adding the binders
    binder_list = []
    for ticket_info in ticket_info_list :
      tls_hash = ticket_info[ 'tls_hash' ]
      binder_finished_key = binder_finished_key_list[ ticket_info_list.index( ticket_info ) ]
      binder = self.compute_binder( tls_hash, binder_finished_key, \
                       truncated_client_hello, \
                       binder_index=ticket_info_list.index( ticket_info ) )
      binder_list.append( { 'binder' : binder } )
    ch_index = self.latest_client_hello_index( )
    self.msg_list[ ch_index ][ 'data' ][ 'extensions' ][ -1 ]\
                 [  'extension_data'] [ 'binders' ] = binder_list


  def update_certificate( self, lurk_cert, server=True ):
    """ build the various certificates payloads

    The resulting Certificate message is then inserted in the payload.
    The case considered are: the server and client Certificates on both
    the TLS server and the TLS client.

    Args:
      lurk_cert (obj) : the LurkCert object
      server (bool) indicates if the lurk_cert is a server (True)
        or a client (False) certificate.
    """
    if self.role == 'server' and server is True or\
       ( self.role == 'client' and server is False ):
      self.msg_list.append( lurk_cert.hs_cert_msg )
    else: #self.role == 'client' and server is True :
      ## inserting the server Certificate
      msg_type_list = self.msg_type_list()
      if 'certificate_verify' not in msg_type_list :
        raise LURKError( 'invalid_handshake', f"CertificateVErify message "\
                                              f"not foun in {msg_type_list}" )
      self.msg_list.insert( msg_type_list.index( 'certificate_verify' ), \
                            lurk_cert.hs_cert_msg )

  def certificate_verify_content( self, role=None ) :
    """ update the handshake with the CertificateVerify """
    if role is None:
      role = self.role
    if role == 'server':
      ctx_string = b'TLS 1.3, server CertificateVerify'
      transcript_h = self.transcript_hash( 'sig' )
    elif role == 'client':
      ctx_string = b'TLS 1.3, client CertificateVerify'
      if self.transcript_r is None:
        transcript_h = self.transcript_hash( 'sig' )
      else:
        transcript_h = self.transcript_hash( 'post_hand_auth_sig' )
    else:
      raise ImplementationError( f"unknown role {self.role}" )
    content = b'\x20' * 64 + ctx_string + b'\x00' + transcript_h
    if self.debug is not None:
      self.debug.trace_val( 'ctx_string', ctx_string )
      self.debug.handle_bin( 'ctx_string', ctx_string )
      self.debug.handle_bin( 'content to be signed', content )
    return content

  def update_certificate_verify( self ) :
    content = self.certificate_verify_content( )
    sig_scheme = SigScheme( self.conf[ 'sig_scheme' ][ 0 ] )
    if sig_scheme.algo in [ 'ed25519', 'ed448' ]:
      signature = self.private_key.sign( content )
    ## ecdsa
    elif sig_scheme.algo == 'ecdsa':
      signature = self.private_key.sign( content, \
                    ECDSA( sig_scheme.hash ) )
    ## rsa
    elif 'rsa' in sig_scheme.algo:
      print( f"--- update_certificate_verify: pad: {sig_scheme.pad}, h: {sig_scheme.hash}" )
      # signature = self.private_key.sign( content, sig_scheme.pad, sig_scheme.hash )
      hasher = Hash( sig_scheme.hash )
      hasher.update( content )
      digest = hasher.finalize( )
      signature = self.private_key.sign( digest, sig_scheme.pad, utils.Prehashed(sig_scheme.hash) )
    else:
      raise LURKError( 'invalid_signature_scheme', f"unknown {sig_scheme.algo}" )

    self.msg_list.append( { 'msg_type' : 'certificate_verify',
                            'data' : { 'algorithm' : sig_scheme.name,
                                       'signature' : signature } } )

  def get_verify_data( self, scheduler, role, transcript_mode ):
    finished_key = scheduler.finished_key( role )
    hmac = HMAC( finished_key, scheduler.tls_hash )
    hmac.update( self.transcript_hash( transcript_mode ) )
    return hmac.finalize()

  def update_finished( self, scheduler ):
    """ updates the finished message

    Only the self.role Finished message is generated.
    More specifically only the TLS client generates the client Finished
    and only the TLS server generates the server Finished.

    get_verify_data is used directly to perform the verification of
    the other Finished message.
    """
    verify_data = self.get_verify_data( scheduler, \
                    role=self.role, transcript_mode=f"{self.role} finished" )
    self.msg_list.append( { 'msg_type' : 'finished',
                            'data' : { 'verify_data' : verify_data } } )


  def get_cipher_suite( self ):
    if self.cipher_suite is None:
      if self.msg_list[ 0 ][ 'msg_type' ] == 'server_hello':
        sh_index = 0
      else:
        ch_index = self.latest_client_hello_index( )
        sh_index = ch_index + 1
      self.cipher_suite = self.msg_list[ sh_index ][ 'data' ][ 'cipher_suite' ]
    return self.cipher_suite

  def get_tls_hash( self ):
    """ returns the instance of the hash function used in the TLS exchange
    """
    return CipherSuite( self.get_cipher_suite() ).get_hash()

  def append_transcript( self, upper_msg_index:int=None  ):
    """ provides a copy of the running transcript

    This transcript contains all or up to msg_index (excluded)
    """
    ## for server we can read it from the configuration file.
    ## this does not considers the case of the client in which
    ## case we will have to read it from the handshake
    ## problem may arise when different format are used between
    ## the client and the server.
#    ctx_struct = { '_certificate_type': self.cert_type }
    ctx_struct = { }
    ## finished message needs the _cipher to be specified
    if 'server_hello' in self.msg_type_list() :
      ctx_struct[ '_cipher' ] = self.get_cipher_suite()

    if upper_msg_index is None:
      msg_list = self.msg_list[ : ]
      del self.msg_list[ : ]
    else:
      msg_list = self.msg_list[ : upper_msg_index ]
      del self.msg_list[ : upper_msg_index ]
    msg_bytes = bytearray()
    for msg in msg_list :
      msg_bytes.extend( tls.Handshake.build( msg, **ctx_struct ) )

    self.transcript.update( msg_bytes )
    transcript = self.transcript.copy()
    h = transcript.finalize()
    return h

  def post_hand_auth_transcript( self ):
    """ post handshake authentication transcript is derived from the transcript
        of the full handshake.
        This is a bit different from append transcript where messages are appened to
        a given transcript.
    """
    if self.transcript_r is None:
      raise ImplementationError( "handshake transcript for the handshake" +\
                                 "has not been finalized - expected to be " +\
                                 "finalized for post hand authentication" )

    transcript = self.transcript.copy()
    msg_bytes = bytearray()
    for msg in self.msg_list :
      msg_bytes.extend( tls.Handshake.build( msg ) )
    transcript.update( msg_bytes )
    del self.msg_list[ : ]
    return transcript.finalize()


  def transcript_hash( self, transcript_type ) -> bytes:
    """ return the Transcript-Hash output for the key schedule

    Performing the hash in the handshake class prevent the
    handshake class to keep all exchanges.
    """

    if self.transcript is None:
      self.transcript = Hash( self.get_tls_hash() )
    upper_msg_index = None

    if transcript_type == 'h' :
      if self.msg_list[ 0 ][ 'msg_type' ] == 'client_hello':
        upper_msg_index = self.latest_client_hello_index( ) + 2
      elif self.msg_list[ 0 ][ 'msg_type' ] == 'server_hello':
        upper_msg_index = 1
      if self.msg_type_list( )[ : upper_msg_index ] not in [ \
        ## 's_init_cert_verify'
        [ 'client_hello', 'server_hello' ], \
        [ 'client_hello', 'server_hello', 'client_hello', 'server_hello' ],
        ## 's_hand_and_app_secret'
        [ 'server_hello' ],
        ]:
        raise LURKError( 'invalid_handshake', f"unexpected handshake {self.msg_list}" )
    ## ClientHello with a HelloRetryRequest, the value of ClientHello1 is
    ## replaced with a special synthetic handshake message of handshake type
    ## "message_hash" containing Hash(ClientHello1).  I.e.,

    ## Transcript-Hash(ClientHello1, HelloRetryRequest, ... Mn) =
    ## Hash(message_hash ||        /* Handshake type */
    ## 00 00 Hash.length  ||  /* Handshake message length (bytes) */
    elif transcript_type == 'sig' :
      if self.msg_type_list( ) not in [ \
        ## 's_init_cert_verify'
        [ 'encrypted_extensions', 'certificate' ],\
        [ 'encrypted_extensions', 'certificate_request', 'certificate' ], \
        ## c_init_client_finished
        [ 'encrypted_extensions', 'certificate_request', 'certificate', \
          'certificate_verify',  'finished', 'certificate' ],
        ## c_init_client_finished
        [ 'end_of_early_data', 'certificate' ],
        [ 'certificate'],
        ['server_hello', 'encrypted_extensions', 'certificate_request',\
         'certificate', 'certificate_verify', 'finished', 'certificate'] ]:
        raise LURKError( 'invalid_handshake', f"unexpected handshake {self.msg_type_list()}" )
##    elif transcript_type == 'finished' :
    elif transcript_type in [ 'client finished', 'server finished' ] :
      if self.msg_type_list( ) not in [ \
        ## 's_init_cert_verify', c_init_client_finished, c_client_finised
        ## without TLS client authentication, the client finished
        ## may be generated with the remaining server Finished
        ## This is a handshake list to consider by the CS for computing
        ## the finished hash.
        [ 'finished' ],\
        ## without TLS client authentication, the client finished
        ## may be generated without additional message after the a_c, a_h
        ## in fact as soon as a_c, a_h are generated by the client
        ## the finished message is not present.
        ## This case happens when the TLS client computes the finished hash
        ## after it has checked the server Finished
        [ ],\
        ## with TLS client authentication. Such list is processed by the CS.
        [ 'finished', 'certificate', 'certificate_verify' ], \
        ## when the TLS client proceeds to the client Finished message,
        ## the server Finished message has been removed to check the
        ## server Finished.
        [ 'certificate', 'certificate_verify' ], \
        ## with certificate authentication (client or server)
        [ 'certificate_verify' ], \
        ## 's_hand_and_app_secret'
        [ 'encrypted_extensions' ], \
        [ 'encrypted_extensions', 'certificate_request' ] ] :
        raise LURKError( 'invalid_handshake', f"unexpected handshake {self.msg_type_list()}" )
    elif transcript_type == 'a' :
      if self.msg_type_list( ) not in [ \
        ## 's_init_cert_verify'
        ## 's_hand_and_app_secret'
        [ 'finished' ], \
        ## c_client_finished
        [ 'server_hello', 'encrypted_extensions', 'certificate_request',\
          'certificate', 'certificate_verify', 'finished' ], \
        [ 'server_hello', 'encrypted_extensions', 'certificate', \
          'certificate_verify', 'finished' ], \
        [ 'server_hello', 'encrypted_extensions', 'finished' ], \
        [ 'encrypted_extensions', 'certificate_request', 'certificate',\
          'certificate_verify', 'finished' ], \
        [ 'encrypted_extensions', 'certificate', 'certificate_verify', 'finished' ], \
        [ 'encrypted_extensions', 'finished' ], \
        ]:
        raise LURKError( 'invalid_handshake', f"unexpected handshake {self.msg_type_list()}" )
    elif transcript_type == 'r' :
      if self.msg_type_list( ) not in [ \
        ## 's_new_ticket
        [ 'finished' ], \
        [ 'certificate', 'certificate_verify', 'finished' ], \
        []] : ## when mutliple s_new_ticket are sent
        raise LURKError( 'invalid_handshake', f"unexpected handshake {self.msg_type_list()}" )
    elif transcript_type == 'e' :
      if self.msg_type_list( ) not in [ \
        ## 's_init_early_secret'
        [ 'client_hello' ], \
        [ 'client_hello', 'server_hello', 'client_hello' ] ] :
        raise LURKError( 'invalid_handshake', f"unexpected handshake {self.msg_type_list()}" )
    elif transcript_type == 'post_hand_auth_sig' :
      if self.msg_type_list( ) not in [ \
        ## c_post_hand_auth
        [ 'certificate_request', 'certificate'] ] :
        raise LURKError( 'invalid_handshake', f"unexpected handshake {self.msg_type_list()}" )
    else:
      raise ImplementationError( f"Unexpected {transcript_type}" )
    ## 'r' may be asked multiple times
    ## in principle, E is expected to send an empty handshake, so that will just work.
    ## However, to support the case when E re-send the handshake messages.
    ## r should be stored and returned instead of being recomputed.
    if transcript_type == 'r' and self.transcript_r is not None:
      transcript = self.transcript_r
    elif transcript_type == 'post_hand_auth_sig' :
      transcript = self.post_hand_auth_transcript( )
    else :
      transcript = self.append_transcript( upper_msg_index )
      if transcript_type == 'r' :
        self.transcript_r = transcript

    if self.debug is not None:
      self.debug.handle_bin( f"Transcript Hash [mode {transcript_type}]", transcript )
    return transcript

  def get_ticket( self, selected_identity:int=None ):
    try:
      ch_index = self.latest_client_hello_index( )
      client_hello_exts = self.msg_list[ ch_index ][ 'data' ][ 'extensions' ]
      pre_shared_key = get_struct(client_hello_exts, 'extension_type', 'pre_shared_key' )
      identities = pre_shared_key[ 'extension_data' ][ 'identities' ]

      if selected_identity is None:
        return identities
      else:
        return identities[ selected_identity ]
    except:
      raise LURKError('invalid_handshake', f"unable to get psk_identity from {identities}" )



from construct.core import *
from construct.lib import *
from construct.debug import *

#import sys
#sys.path.insert(0, '/home/emigdan/gitlab/pytls13/src/')
import pytls13.struct_tls13 as tls
#sys.path.insert(0, '/home/emigdan/gitlab/pylurk.git/src')
import pylurk.tls13.struct_tls13 as lurk

TLS13Version = Enum( BytesInteger(1), 
  v1 = 1
)

TLS13Type = Enum( BytesInteger(1), 
  capabilities = 0, 
  ping = 1,
  s_init_cert_verify = 2,
  s_new_ticket = 3,
  s_init_early_secret = 4,
  s_hand_and_app_secret = 5,
  c_init_client_finished = 6,  
  c_post_hand_auth = 7,        
  c_init_client_hello = 8,    
  c_client_hello = 9,         
  c_server_hello = 10,         
  c_client_finished = 11,      
  c_register_tickets =12     
)

TLS13Status = Enum( BytesInteger(1), 
  request = 0,
  success = 1,
  invalid_status = 15,
  undefined_error = 2,
  invalid_format = 3,
  invalid_secret_request = 4, 
  invalid_session_id = 5,
  invalid_handshake = 6, 
  invalid_freshness = 7, 
  invalid_ephemeral = 8, 
  invalid_psk = 9, 
  invalid_certificate = 10,
  invalid_cert_type = 12,
  invalid_type = 11,
)

## Common structures
SecretRequest = BitStruct(
  "b"   / Flag, 
  "e_s" / Flag,
  "e_x" / Flag,
  "h_c" / Flag,
  "h_s" / Flag,
  "a_c" / Flag,
  "a_s" / Flag,
  "x"   / Flag,
  "r"   / Flag,
  "reserved" / Const(0,BitsInteger(7))
)


SecretType = Enum ( BytesInteger(1),
  b = 0,
  e_s = 1,
  e_x = 2,
  h_c = 3,
  h_s = 4,
  a_c = 5, 
  a_s = 6,
  x = 7, 
  r = 8
)

Secret = Struct(
  '_name' / Computed('Secret'),
  'secret_type' / SecretType, 
  'secret_data' / Prefixed( BytesInteger(1), GreedyBytes )
)


HandshakeList = Switch( this._type, 
  { 's_init_early_secret' : Sequence( lurk.HSClientHello ),
    's_init_cert_verify' : Select( \
      Sequence( tls.HSClientHello,
                lurk.HSServerHello, 
                tls.HSEncryptedExtensions, 
                tls.HSCertificateRequest ),
      Sequence( lurk.HSClientHello, 
                lurk.HSServerHello, 
                tls.HSEncryptedExtensions )),  
    's_new_ticket' : Select( \
      Sequence( tls.HSCertificateVerify, 
                tls.HSFinished ),
      Sequence( tls.HSFinished ) ),
    's_hand_and_app_secret' : Select(\
      Sequence( lurk.HSServerHello, 
                tls.HSEncryptedExtensions, 
                tls.HSCertificateRequest ),
      Sequence( lurk.HSServerHello, tls.HSEncryptedExtensions ) ),  
    'c_init_client_finished' : Select( \
       ## if cert_request + server certificate + cert_verify ):
       ##   -> client cert 
       ## elif cert cert_verify but no cert_request
       ##  -> nothing 
       ## elif no cert, no cert_verify, no cert request
       ##  -> end of earlydata yes / no
       Sequence( tls.HSClientHello, 
                 tls.HSServerHello, 
                 tls.HSEncryptedExtensions, 
                 tls.HSCertificateRequest, 
                 tls.HSCertificateVerify, 
                 tls.HSFinished), 
       Sequence( tls.HSClientHello, 
                 tls.HSServerHello, 
                 lurk.HSClientHello, 
                 tls.HSEncryptedExtensions, 
                 tls.HSCertificateRequest, 
                 tls.HSCertificateVerify, 
                 tls.HSFinished),
       Sequence( tls.HSClientHello, 
                 tls.HSServerHello, 
                 tls.HSEncryptedExtensions, 
                 tls.HSCertificateVerify, 
                 tls.HSFinished), 
       Sequence( tls.HSClientHello, 
                 tls.HSServerHello, 
                 lurk.HSClientHello, 
                 tls.HSEncryptedExtensions, 
                 tls.HSCertificateVerify, 
                 tls.HSFinished),
       Sequence( tls.HSClientHello, 
                 tls.HSServerHello, 
                 tls.HSEncryptedExtensions, 
                 tls.HSFinished, 
                 tls.HSEndOfEarlyData),  
       Sequence( tls.HSClientHello, 
                 tls.HSServerHello, 
                 tls.HSEncryptedExtensions, 
                 tls.HSFinished) ), 
    'c_post_hand_auth' : Sequence( tls.HSCertificateRequest), 
    'c_init_client_hello' : Sequence( lurk.HSClientHello ),
    'c_server_hello' : Sequence( lurk.HSServerHello ),
    'c_client_finished' : Select( \
      ## case when serverhello is provided ( there is no c_server_hello exchange)
      Sequence( tls.HSServerHello, 
                tls.HSEncryptedExtensions, 
                tls.HSCertificateRequest, 
                tls.HSCertificateVerify, 
                tls.HSFinished ),
      Sequence( tls.HSServerHello, 
                tls.HSEncryptedExtensions, 
                tls.HSCertificateVerify, 
                tls.HSFinished ),
      Sequence( tls.HSServerHello, 
                tls.HSEncryptedExtensions, 
                tls.HSFinished, 
                tls.HSEndOfEarlyData),
      Sequence( tls.HSServerHello, 
                tls.HSEncryptedExtensions, 
                tls.HSFinished ),
      Sequence( tls.HSEncryptedExtensions, 
                tls.HSCertificateRequest, 
                tls.HSCertificateVerify, 
                tls.HSFinished ),
      Sequence( tls.HSEncryptedExtensions, 
                tls.HSCertificateVerify, 
                tls.HSFinished ),
      Sequence( tls.HSEncryptedExtensions, 
                tls.HSFinished, 
                tls.HSEndOfEarlyData),
      Sequence( tls.HSEncryptedExtensions, 
                tls.HSFinished ) ),
  }, default=Error
)

Freshness = Enum( BytesInteger(1),
    sha256 = 0,
    sha384 = 1,
    sha512 = 2,
    null = 255,
)


EphemeralMethod = Enum ( BytesInteger(1),
  no_secret = 0,
  e_generated = 1,
  cs_generated = 2
)

SharedSecret = Struct(
  '_name' / Computed('SharedSecret'),
  'group' / tls.NamedGroup, 
  'shared_secret' / Switch(this.group, 
  { 'secp256r1' : Bytes(32),
    'secp384r1' : Bytes(48),
    'secp521r1' : Bytes(66),
    'x25519' : Bytes(32),
    'x448' : Bytes(56)
  }, Error)
)

Ephemeral = Struct(
  '_name' / Computed('Ephemeral'),
  '_status' / Computed( this._._status ),
  'method' / EphemeralMethod,
#  Probe(),
  'key' / Switch(this.method,
    { 'e_generated' : Switch(this._status, {
         'request' : Prefixed(BytesInteger(2), SharedSecret),
         'success' : Const(b'')
        }, Error),
      'cs_generated' : Switch(this._status, {
         'request' : Const( b'' ), 
         'success' : Prefixed(BytesInteger(2), tls.KeyShareEntry),
        }, Error),
      'no_secret' : Switch(this._status, {
         'request' : Const( b'' ), 
         'success' : Const( b'' ),
    }, Error)
  } )
)


CertType = Enum ( BytesInteger(2),
  zlib = 1, 
  brotli = 2, 
  zstd = 3,
  no_certificate = 128,
  finger_print = 129,
  uncompressed = 130
)

FingerPrintCertificateEntry = Struct(
  '_name' / Computed('FingerPrintCertificateEntry'),
  'finger_print' / Bytes( 4 ),
  'extensions' / Prefixed( BytesInteger(2), GreedyRange( tls.Extension ) )
)

FingerPrintCertificate = Struct(
  '_name' / Computed('FingerPrintCertificate'),
  'certificate_request_context' / Prefixed( BytesInteger(1), GreedyBytes),
  'certificate_list' / Prefixed(BytesInteger(3), GreedyRange(FingerPrintCertificateEntry))
)

Cert = Struct(
  '_name' / Computed('Cert'),
  'cert_type' /  CertType,
#  Probe( this.cert_type ),
  'certificate' / Switch( this.cert_type, {
    'zlib' : tls.CompressedCertificate, 
    'brotli' : tls.CompressedCertificate,
    'zstd' : tls.CompressedCertificate,
    'no_certificate' : Const( b'' ), 
    'finger_print' : FingerPrintCertificate,
    'uncompressed' : tls.Certificate
    }
  )
)

Tag = BitStruct(
  "last_exchange"   / Flag, 
  "reserved" / Const(0,BitsInteger(7))
)

#PSKMethod = Enum ( BytesInteger(1),
#  e = 1,
#  cs = 2
#)

TLSHash = Enum( BytesInteger(1),
    sha256 = 0,
    sha384 = 1,
    sha512 = 2,
)

PSKType = Enum( BytesInteger(1),
    external = 0,
    resumption = 1,
)


PskIdentityMetadata = Struct(
  '_name' / Computed('PskIdentityMetadata'),
  'identity_index' / BytesInteger( 2 ),
  'tls_hash' / TLSHash,
  'psk_type' / PSKType, 
  'psk_bytes' / Prefixed( BytesInteger(2), GreedyBytes),
)



SInitEarlySecretRequest = Struct(
  '_name' / Computed('SInitEarlySecretRequest'),
  '_type' / Computed(this._._type),
  'session_id' / Bytes( 4 ),
  'freshness' / Freshness, 
  'selected_identity' / BytesInteger( 2 ), 
  'handshake' / Prefixed( BytesInteger(4), HandshakeList ),
  'secret_request' / SecretRequest,
)


SInitEarlySecretResponse = Struct(
  '_name' / Computed('SInitEarlySecretResponse'),
  '_type' / Computed(this._._type),
  'session_id' / Bytes( 4 ),
  'secret_list' / Prefixed( BytesInteger(2), GreedyRange( Secret ) )
)


SInitCertVerifyRequest = Struct(
  '_name' / Computed('SInitCertVerifyRequest'),
  '_type' / Computed(this._._type),
  '_status' / Computed('request'),
  'tag' / Tag,
  'session_id' / Switch( this.tag.last_exchange,
    { True : Const( b'' ), 
      False : Bytes( 4 ),
    }  
  ),
  'freshness' / Freshness,
  'ephemeral' / Ephemeral, 
  'handshake' / Prefixed( BytesInteger(4), HandshakeList ), 
  'certificate' / Cert, 
  'secret_request' / SecretRequest,
  'sig_algo' / tls.SignatureScheme,
)

SInitCertVerifyResponse = Struct(
  '_name' / Computed('SInitCertVerifyResponse'),
  '_type' / Computed(this._._type),
  '_status' / Computed('success'),
  'tag' / Tag,
  'session_id' / Switch( this.tag.last_exchange,
    { True : Const(b''), 
      False : Bytes(4),
    }  
  ),
  'ephemeral' / Ephemeral, 
  'secret_list' / Prefixed( BytesInteger(2), GreedyRange( Secret ) ),
  'signature' / Prefixed( BytesInteger(2), GreedyBytes )
)


SHandAndAppRequest = Struct(
  '_name' / Computed('SHandAndAppRequest'),
  '_type' / Computed(this._._type),
  '_status' / Computed('request'),
  'tag' / Tag,
  'session_id' / Bytes( 4 ),
  'ephemeral' / Ephemeral, 
  'handshake' / Prefixed( BytesInteger(4), HandshakeList ), 
  'secret_request' / SecretRequest
)
  
SHandAndAppResponse = Struct(
  '_name' / Computed('SHandAndAppResponse'),
  '_status' / Computed('success'),
  'tag' / Tag,
  'session_id' / Bytes(4) , 
  'ephemeral' / Ephemeral, 
  'secret_list' / Prefixed( BytesInteger(2), GreedyRange( Secret ) )
)


SNewTicketRequest = Struct(
  '_name' / Computed('SNewTicketRequest'),
  '_type' / Computed(this._._type),
  'tag' / Tag,
  'session_id' / Bytes(4), 
  'handshake' / Prefixed( BytesInteger(4), HandshakeList),
  'certificate' / Cert, 
  'ticket_nbr' / BytesInteger(1), 
  'secret_request' / SecretRequest,
)

SNewTicketResponse = Struct(
  '_name' / Computed('SNewTicketResponse'),
  'tag' / Tag,
  'session_id' / Bytes(4),
  'secret_list' / Prefixed( BytesInteger(2), GreedyRange( Secret ) ),
  'ticket_list' / Prefixed( BytesInteger( 2 ), GreedyRange( tls.NewSessionTicket ) )
)


## LURK request / response structures on the TLS client

CInitClientFinishedRequest = Struct(
  '_name' / Computed('CInitClientFinishedRequest'),
  '_type' / Computed('c_init_client_finished'),
  '_status' / Computed('request'),
  'tag' / Tag,
  'session_id' / Switch( this.tag.last_exchange,
    { True : Const( b'' ), 
      False : Bytes( 4 ),
    }  
  ),
  'handshake' / Prefixed( BytesInteger(4), HandshakeList ), 
  'server_certificate' / Cert, 
  'client_certificate' / Cert, 
  'freshness' / Freshness,
  'ephemeral' / Ephemeral, 
  'psk' / Prefixed( BytesInteger(2), GreedyBytes ), 
)

CInitClientFinishedResponse = Struct(
  '_name' / Computed('CInitClientFinishedResponse'),
  '_type' / Computed('c_init_client_finished'),
  '_status' / Computed('success'),
  'tag' / Tag,
  'session_id' / Switch( this.tag.last_exchange,
    { True : Const(b''), 
      False : Bytes(4),
    }  
  ),
  'signature' / Prefixed( BytesInteger(2), GreedyBytes )
)

CPostHandAuthRequest = Struct(
  '_name' / Computed('PostHandRequest'),
  '_type' / Computed('c_post_hand_auth'),
  'tag' / Tag,
  'session_id' / Bytes(4), 
  'handshake' / Prefixed( BytesInteger(4), HandshakeList ),
  'certificate' / Cert
)

CPostHandAuthResponse = Struct(
  '_name' / Computed('PostHandResponse'),
  'tag' / Tag,
  'session_id' / Bytes(4), 
  'signature' / Prefixed( BytesInteger(2), GreedyBytes )
)


CInitClientHelloRequest = Struct(
  '_name' / Computed('CInitClientHelloRequest'),
  '_type' / Computed('c_init_client_hello'),
  'session_id' / Bytes( 4 ),
  'handshake' / Prefixed( BytesInteger(4), HandshakeList), 
  'freshness' / Freshness, 
  'psk_metadata_list' / Prefixed( BytesInteger(2), GreedyRange( PskIdentityMetadata ) ), 
  'secret_request' / SecretRequest
)


CInitClientHelloResponse = Struct(
  '_name' / Computed('CInitClientHelloResponse'),
  '_status' / Computed('success'),
  'session_id' / Bytes( 4 ), 
  'ephemeral_list' / Prefixed( BytesInteger(2), GreedyRange( Ephemeral ) ),
  'binder_key_list' / Prefixed( BytesInteger(2), GreedyRange( Secret ) ),
  'secret_list' / Prefixed( BytesInteger(2), GreedyRange( Secret ) )
) 


CServerHelloRequest = Struct(
  '_name' / Computed('CServerHelloRequest'),
  '_type' / Computed('c_server_hello'),
  '_status' / Computed('request'),
  'session_id' / Bytes( 4 ),
  'handshake' / Prefixed( BytesInteger(4), HandshakeList), 
  'ephemeral' / Ephemeral, 
)

CServerHelloResponse = Struct(
  '_name' / Computed('CServerHelloResponse'),
  '_type' / Computed('c_server_hello'),
  '_status' / Computed('success'),
  'session_id' / Bytes( 4 ),
  'secret_list' / Prefixed( BytesInteger(2), GreedyRange( Secret ) )
)

CClientFinishedRequest = Struct(
  '_name' / Computed('CClientFinishedRequest'),
  '_type' / Computed('c_client_finished'),
  '_status' / Computed('request'),
  'tag' / Tag,
  'session_id' / Bytes( 4 ),
  'handshake' / Prefixed( BytesInteger(4), HandshakeList ), 
  'server_certificate' / Cert, 
  'client_certificate' / Cert,
  'secret_request' / SecretRequest
)

CClientFinishedResponse = Struct(
  '_name' / Computed('CClientFinishedResponse'),
  '_type' / Computed('c_client_finished'),
  '_status' / Computed('success'),
  'tag' / Tag,
  'session_id' / Bytes( 4 ),
  'signature' / Prefixed( BytesInteger(2), GreedyBytes ),
  'secret_list' / Prefixed( BytesInteger(2), GreedyRange( Secret ) )
)

CRegisterTicketsRequest = Struct(
  '_name' / Computed('CRegisterTicketRequest'),
  '_type' / Computed('c_register_tickets'),
  '_status' / Computed('request'),
  'tag' / Tag,
  'session_id' / Bytes( 4 ),
  'ticket_list' / Prefixed( BytesInteger(2), GreedyRange( tls.NewSessionTicket ) ), 
)

CRegisterTicketsResponse = Struct(
  '_name' / Computed('CRegisterTicketRequest'),
  '_type' / Computed('c_register_tickets'),
  '_status' / Computed('success'),
  'tag' / Tag,
  'session_id' / Bytes( 4 ),
)

############  LURKErrorPayload

ErrorPayload = Struct (
  'lurk_state' / Bytes(4), \
)


TLS13Payload = Switch(this._type, 
  { 's_init_early_secret' : Switch( this._status,
       { 'request' : SInitEarlySecretRequest, 
         'success' : SInitEarlySecretResponse, 
        }, default=ErrorPayload), 
    's_init_cert_verify' : Switch( this._status,
       { 'request' : SInitCertVerifyRequest, 
         'success' : SInitCertVerifyResponse, 
        }, default=ErrorPayload), 
    's_hand_and_app_secret' : Switch( this._status,
       { 'request' : SHandAndAppRequest, 
         'success' : SHandAndAppResponse, 
        }, default=ErrorPayload), 
    's_new_ticket' : Switch( this._status,
       { 'request' : SNewTicketRequest, 
         'success' : SNewTicketResponse, 
        }, default=ErrorPayload), 
    'c_init_client_finished' : Switch( this._status,
       { 'request' : CInitClientFinishedRequest, 
         'success' : CInitClientFinishedResponse, 
        }, default=ErrorPayload), 
    'c_post_hand_auth' : Switch( this._status,
       { 'request' : CPostHandAuthRequest, 
         'success' : CPostHandAuthResponse, 
        }, default=ErrorPayload), 
    'c_init_client_hello' : Switch( this._status,
       { 'request' : CInitClientHelloRequest, 
         'success' : CInitClientHelloResponse, 
        }, default=ErrorPayload), 
    'c_server_hello' : Switch( this._status,
       { 'request' : CServerHelloRequest, 
         'success' : CServerHelloResponse, 
        }, default=ErrorPayload), 
    'c_client_finished' : Switch( this._status,
       { 'request' : CClientFinishedRequest, 
         'success' : CClientFinishedResponse, 
        }, default=ErrorPayload), 
    'c_register_tickets' : Switch( this._status,
       { 'request' : CRegisterTicketsRequest, 
         'success' : CRegisterTicketsResponse, 
        }, default=ErrorPayload), 
  }, default=ErrorPayload
)






import logging
import socketserver
import selectors
import time
#import ssl
#import ssl.SSLError

from pylurk.struct_lurk import LURKMessage, LURKHeader
from pylurk.lurk.lurk_lurk   import LURKError, ImplementationError, ConfigurationError, LurkExt
####import sys
###sys.path.insert(0, '/home/emigdan/gitlab/pylurk.git/src/')
import pylurk.tls13.lurk_tls13
import pylurk.debug
#from pylurk.tls13.lurk_tls13  import Tls13Ext, TicketDB, SessionDB

MINIMUM_PACKET_SIZE = 16
HEADER_SIZE = 12

def logger( conf, __name__ ):
  """ returns a looger object """
  try:
    log_file = conf[ 'log' ]
  except KeyError :
    log_file = './crypto_service.log'
  logger = logging.getLogger( __name__ )
  FORMAT = "[%(asctime)s : %(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s"
  logger.setLevel( logging.DEBUG )
  logging.basicConfig( filename=log_file, format=FORMAT )
  return logger

class BaseCryptoService:
  def __init__( self, conf ):
    self.con_type = 'lib_cs'
    self.conf = conf

    for ext in self.conf[ 'enabled_extensions' ]:
      if ext == ( 'lurk', 'v1' ):
        self.lurk = LurkExt( self.conf  )
      elif ext == ( 'tls13', 'v1' ):
        self.ticket_db = pylurk.tls13.lurk_tls13.TicketDB()
        self.session_db = pylurk.tls13.lurk_tls13.SessionDB()

        self.debug = None
        debug_conf = self.conf[ ext ][ 'debug' ]
        debug = pylurk.debug.Tls13Debug( debug_conf )
        if debug.trace is True or debug.test_vector is True:
          self.debug = debug

        self.tls13 = pylurk.tls13.lurk_tls13.Tls13Ext( self.conf[ ext ],\
                       ticket_db=self.ticket_db,\
                       session_db=self.session_db,\
                       debug=self.debug )

#        raise ValueError( )
    try:
      self.lurk_state = self.lurk.lurk_state
    except NameError:
      raise ConfigurationError( "( 'lurk', 'v1' ) MUST be enabled" )
    self.logger = logger( self.conf, __name__ )

  def serve( self, req_bytes ) -> bytes :
    """ returns a binary response """
    if len( req_bytes ) < MINIMUM_PACKET_SIZE :
      return b''

    try :
      header = LURKHeader.parse( req_bytes[ : HEADER_SIZE ] )
      resp = { 'designation' : header[ 'designation' ],
               'version' : header[ 'version' ],
               'type' : header[ 'type' ],
               'status' : None,
               'id' : header[ 'id' ],
               'payload' : {} }
      ext = ( header[ 'designation' ], header[ 'version' ] )
      if ext  not in self.conf[ 'enabled_extensions' ] :
        raise LURKError( 'invalid_extension' , f"{ext}" )
      if header[ 'type' ] not in self.conf[ ext ] [ 'type_authorized' ] :
        raise LURKError( 'invalid_type' , f"{header[ 'type' ]}" )
      if header[ 'status' ] != 'request' :
        raise LURKError( 'invalid_status' , f"{header[ 'status' ]}" )
      req = LURKMessage.parse( req_bytes )
      if self.debug is not None:
        self.debug.handle_lurk_msg( req )
      if ext == ( 'lurk', 'v1' ):
        payload = self.lurk.payload_resp( req )
      elif ext == ( 'tls13', 'v1' ):
        payload = self.tls13.payload_resp( req )
      else:
        raise LURKError( 'invalid_extension' , f"{ext}" )
      resp[ 'payload' ] = payload
      resp[ 'status' ] = 'success'
      if self.debug is not None:
        self.debug.handle_lurk_msg( resp)
      return LURKMessage.build( resp )
    except Exception as e :
      if isinstance( e, LURKError ):
        resp[ 'status' ] = e.status
        self.logger.info( f"{str( e )} - {e.message}" )
      else:
        if isinstance( e, ( ImplementationError, ConfigurationError ) ):
          self.logger.error( f"{e.status} : {e.message}" )
        else:
          self.logger.exception( str( e ) )
        resp[ 'status' ] = 'undefined_error'
      resp[ 'payload' ] = { 'lurk_state' : self.lurk_state }
      return LURKMessage.build( resp )


class NonPersistentTCPHandler( socketserver.BaseRequestHandler):
  """ the NonPersistentTCPHandler assumes that the packet it handles
      contains only one LURK request.
      TCP ensures all bytes are transported but a TCP session is
      established for every LURK request.
  """
  def handle( self ):
    req_bytes = self.request.recv(4096)
    resp = self.server.cs.serve( req_bytes )
    self.request.sendall( resp )

class NonPersistentTCPCryptoService( socketserver.TCPServer):

  def __init__( self, conf  ):
    """ Non Persistent TCP service

    The non persistent TCP Crypto Service handles one TCP session per
    packet.
    The NonPersistentTCP server extends the TCP calls to instantiate
    a crypto service as self.cs.
    """

    self.con_type = 'tcp'
    self.conf = conf
    self.cs = BaseCryptoService( self.conf )
    server_address = self.get_server_address_from_conf( )
    super().__init__( server_address, NonPersistentTCPHandler,\
                       bind_and_activate=True )

  def get_server_address_from_conf( self ):
    """ return host and port from the configuration file """

    key_list = self.conf[ 'connectivity' ].keys()
    if 'type' in key_list:
      cs_type = self.conf[ 'connectivity' ][ 'type' ]
      if cs_type != self.con_type:
        raise ConfigurationError( f"unexpected type {cs_type} for "\
          f"{self.__class__.__name__}. Expecting '{self.con_type}'." )
    else:
      raise ConfigurationError( f"Cannot find type in configuration "\
        f"{self.__class__.__name__}. Expecting 'tcp'." )
    host = None
    if 'fqdn' in key_list:
      fqdn = self.conf[ 'connectivity' ][ 'fqdn' ]
      if fqdn not in [ None, '' ]:
        host = fqdn
    if host is None and 'ip' in key_list:
      host = self.conf[ 'connectivity' ][ 'ip' ]
    if 'port' in key_list:
      port = self.conf[ 'connectivity' ][ 'port' ]
    else:
      raise ConfigurationError( f"Cannot find port in configuration "\
        f"{self.__class__.__name__}." )
    return ( host, port )
#    with socketserver.TCPServer((host, port), MyTCPHandler) as server:
#        # Activate the server; this will keep running until you
#        # interrupt the program with Ctrl-C
#        server.serve_forever()



#class BaseTCPServer(TCPServer):
class PersistentTCPServer( socketserver.TCPServer ):

  def __init__(self, server_address, RequestHandlerClass,\
             bind_and_activate=True):
    """Basic TCP Server

    The main difference with the TCPServer class is that TCPServer
    class accepts a TCP session for a request, process the request
    and close the TCP session. The advantage is that it prevents
    management or tracking of unused TCP session with a timeout
    for example. The downside is that it also rpevents a TCP
    session to be used for multiple requests.

    This class modify the TCPServer class by 1) not shuting down and
    closing the socket after the initial request has been treated by
    the RequestHandlerClass. 2) listen to events happening on the
    listening socket (self.socket) as well as those accepted sockets
    (self.accept()). The latest are used when further requests are
    sent over the established TCP session. 3) sockets needs to be
    managed and eventually closed when timeout occurs.
    """

    self.allow_reuse_address = True
    super().__init__( server_address, RequestHandlerClass,\
                      bind_and_activate )
#    if self.connection_type == 'tcp+tls':
#        context = self.conf.get_tls_context()
#        self.socket = context.wrap_socket(self.socket, server_side=True)
    ## selector listen to sockets available to the server.
    ## there are 2 kind of sockets.
    ## 1. The server socket ( self.socket ) that is contacted
    ##   by lurk client to establish a communication socket.
    ##   This socket is designated as "accept" (see data value).
    ## 2. sockets establsihed between each lurk client and the
    ##   server. These sockets are used to transmit the LURK
    ##   exchanges and are designated as "established".
    ##
    ## All these sockets are listen for the READ event.
    self.selector = selectors.DefaultSelector()
    ## registering the "accept" socket
    self.selector.register(fileobj=self.socket, \
                           events=selectors.EVENT_READ, \
                           data="accept")\
    ## the time after wich a socket is closed in case
    ## of non activity
    self.fd_timeout = 30
    ## inactivity time associated to each socket (file descriptor)
    ## socket are identied by their file descriptor number
    self.fd_time = {}
    ## list of fd (sockets) being treated
    self.fd_busy = {}

  def shutdown_request(self, request):
    """ actions after the RequestHandlerClass is called.

    TCPServer closes the socket used by the handler. This results in
    having socket being used for a single transaction. As we are
    looking to be able to re-use a socket that has been accepted
    for further transactions, the socket needs to be left open.
    The current function prevents shutingdown and closing the socket.

    Args:
        request: a socket object.
    """
    pass


  def serve_forever(self, poll_interval=None):
    """ serves incoming request

    Handles requests until shutdown

    The main difference with the original BaseServer function
    is that the BaseServer function is using non persitent TCP
    sessions which means that only the server socket is
    being monitored.
    In the Base server every incoming TCP session waits for:

    1. the creation of a new socket (sock = self.sock.accept()).
    2. the response to be sent on this newly established socket
    3. that newly established socket being closed.

    In our case we need to monitore all persistent session that
    is:

    1. TCP session that are in an establishment
    2. Established TCP sessions

    An incoming TCP session message take sthe following path:

    1. Identify which socket is associated to the message
    2. If a new socket needs to be created creates it and
      monitor further actions.
    3. For every socket on whic a message has been sent
      Treat that message.

    The main difference is that for any action, the socket
    needs to be specified.
    """

    print("staring serve_forever")
    self._BaseServer__is_shut_down.clear()
    previous_time = 0
    try:
      while not self._BaseServer__shutdown_request:
        ## listen to ready files and returns ( key, event )
        ## with poll_intervall it block until a file is ready.
        ## selector.select returns the following list:
        ## [(SelectorKey(fileobj=<socket.socket fd=3,
        ##   family=AddressFamily.AF_INET,
        ##   type=SocketKind.SOCK_STREAM, proto=0,
        ##   laddr=('127.0.0.1', 9402)>, fd=3, events=1,
        ##   data='accept'), 1)]
        events = self.selector.select( poll_interval )
        for selector_key, event in events:
          ## ensure the server can be shutdown
          if self._BaseServer__shutdown_request:
            break
          ## check fd (socket) is already being treated
          try:
            self.fd_busy[ selector_key.fileobj.fileno() ]
          except KeyError:
            self._handle_request_noblock(selector_key, event)
            self.service_actions() ## pass
        ## closing sockets that have been opened for
        ## more than timeout
        current_time = time.time()
        if current_time - previous_time > 1:
          previous_time = current_time
          ##
          for fd in self.selector.get_map():
            key = self.selector._fd_to_key[ fd ]
            try:
              delta_time = current_time - self.fd_time[ fd ]
              if delta_time > self.fd_timeout and\
                 key.data == 'establish':
                self.close_request( key.fileobj )
            except KeyError as e:
              ## time of self.socket is not monitored
              ## while it triggers events
              continue
    finally:
      self._BaseServer__shutdown_request = False
      self._BaseServer__is_shut_down.set()

  def _handle_request_noblock(self, selector_key, event):
    """ Handle one request

    The only change from the original TCP is that we indicate
    the ( socket, event ) to be passes to get_request.

    In the BaseServer, such information is inferred from the
    server_socket. In our case, server_socket is only one
    possibility and established socket MUST be considered.

    """
    try:
      request, client_address = self.get_request(selector_key, event)
#      print( f"DEBUG: request {request}" )
    except ( TypeError, OSError ):
      return
    else:
      if self.verify_request(request, client_address):
        try:
#          print( f"DEBUG: processing_request" )
          self.process_request(request, client_address)
        except Exception:
          self.handle_error(request, client_address)
          self.shutdown_request(request)
        except:
          self.shutdown_request(request)
          raise
      else:
        self.shutdown_request(request)


  def get_request(self, selector_key, event):
    """Provides connectivity information to the RequestHandlerClass

    Returns the appropriated socket (request) and address
    (client_address) to the RequestHandlerClass. The parameters are
    passed via the finish_request method.
    serve_forever() --> _handle_request_noblock() -->
    process_request(self, request, client_address) -->
    finish_request(self, request, client_address)

    Args:
       selector_key: SelectorKey object (fileobj, fd, events, data).
           It is returned by the selector.select()
       event:
    """
    if selector_key.data == "accept":
      request, client_address = self.socket.accept()
      ## the establishement does not indicate there is some data
      ## to be received. As aresult, one needs to put the socket
      ## in a non blocking state to prevent the socket to block
      ## waiting to receive any data.
      request.setblocking(False)
#      if self.connection_type == 'tcp+tls':
#          context = self.conf.get_tls_context()
#          request = context.wrap_socket(request, server_side=True,\
#                                        do_handshake_on_connect=False)
      self.selector.register(fileobj=request, \
                             events=selectors.EVENT_READ,\
                             data="establish")
#      ## registering the last tim eactivity o fthe socket
#      self.fd_time[request.fileno] = time.time()
    elif selector_key.data == "establish":
      request = selector_key.fileobj
      client_address = request.getpeername()
    ## registering the last time activity o fthe socket
    self.fd_time[request.fileno] = time.time()
    return request, client_address

  def close_request(self, request):
    self.selector.unregister(request)
    request.close()

MAX_ATTEMPTS = 3
class PersistentTCPHandler( socketserver.BaseRequestHandler):
  """
  """

  def handle(self):
    """ treat the request
    From https://docs.python.org/3.5/library/socketserver.html

    This function must do all the work required to service a
    request. The default implementation does nothing. Several
    instance attributes are available to it; the request is
    available as self.request; the client address as
    self.client_address; and the server instance as self.server,
    in case it needs access to per-server information.

    The type of self.request is different for datagram or stream
    services.  For stream services, self.request is a socket object.


    """
    ## check the sicket is not being treated
    fileno = self.request.fileno()
    if fileno in self.server.fd_busy.keys():
      return
    self.server.fd_busy[ fileno ] = time.time()

#    try:
#      self.server.fd_busy[ self.request.fileno() ]
##      return
#    except KeyError:
#
##      self.server.fd_busy[ self.request.fileno() ] = time.time()
    try:
#      print( f"DEBUG: receiving bytes" )
      ## we need to consider the additional len
      ## of the payload (4 bytes)
      req_bytes = self.request.recv( HEADER_SIZE + 4 )
#      print( f"DEBUG: handle : bytes_recv : {req_bytes}" )
    ## BlockingIOError error 11 happens when a socket
    ## in a non blocking mode does not have any data.
    ## The reason we considered non blocking mode is when
    ## a client just connect, without sending requests.
    ## We do not want the server be blocked until data
    ## is being sent.
    ## Maybe we could set a timer to recv.
    except BlockingIOError as e :
#      print( f"DEBUG: Excpetion {type(e)} {e}" )
      del self.server.fd_busy[ self.request.fileno() ]
      return
    else:
      if req_bytes == b'':
        return
      ## the payload len defines teh remaining bytes to be receievd.
      payload_len = int.from_bytes( req_bytes[ -4 : ], byteorder='big' )
#      header = LURKHeader.parse( req_bytes )
      #bytes_nbr = header[ 'length' ]
      attempt_nbr = 0
      while len( req_bytes ) < payload_len + HEADER_SIZE + 4:
        try:
          remaining_bytes = payload_len + HEADER_SIZE + 4 - len( req_bytes )
          req_bytes += self.request.recv( min( remaining_bytes, 4096) )
#        except ssl.SSLError as err:
#          if err.args[0] == ssl.SSL_ERROR_WANT_READ:
#            select([self.request], [], [])
#          elif err.args[0] == ssl.SSL_ERROR_WANT_WRITE:
#            select([self.request], [], [])
#          else:
#            raise
        except BlockingIOError:
          attempt_nbr += 1
          if attempt_nbr == MAX_ATTEMPTS:
            return
          ## currently we just sleep for 1 sec, but we may
          ## also wait for the socket to raise a read event.
          time.sleep( 1 )
         #   self.server.selector.select([ self.request ], [], [], 5)
      attempt_nbr = 0
#      print( f"DEBUG: CS: request bytes: {req_bytes}" )
      while attempt_nbr <= MAX_ATTEMPTS:
        try:
          resp = self.server.cs.serve( req_bytes )
##         self.request.sendall(self.server.byte_serve(bytes_recv))
          self.request.sendall( resp )
#          print( f"DEBUG: CS: resp: {resp}" )
          break
#        except ssl.SSLError as err:
#          if err.args[0] == ssl.SSL_ERROR_WANT_WRITE:
#            select([], [ self.request ], [])
#          else:
#            raise
        except BlockingIOError:
          select([], [ self.request ], [])
        if attempt_nbr == MAX_ATTEMPTS:
          return
          raise ImplementationError(attempt_nbr, "Reading Header" +\
                  "attempts exceeds MAX_ATTEMPTS " +\
                  "= %s"%MAX_ATTEMPTS +\
                  "Lurk Header not read")
      del self.server.fd_busy[self.request.fileno()]
      return

class PersistentTCPCryptoService( PersistentTCPServer ):

  def __init__( self, conf  ):
    """

    The Persitent TCP Crypto Service handles persistent TCP sessions
    The TCP server extends the TCP calls to instantiate
    a crypto service as self.cs.
    """

    self.con_type = 'persistent_tcp'
    self.conf = conf
    self.cs = BaseCryptoService( self.conf )
    server_address = NonPersistentTCPCryptoService.get_server_address_from_conf( self )
#    super().__init__( server_address, PersistentTCPHandler,\
#                       bind_and_activate=True )
    PersistentTCPServer.__init__( self, server_address, \
                                  PersistentTCPHandler, \
                                  bind_and_activate=True )


def get_cs_instance( conf ):
  """ returns appropriated Cryptographic Service instance from conf

  """
  con_type = conf[ 'connectivity' ][ 'type' ]
  if con_type == 'lib_cs':
    cs = BaseCryptoService( conf )
  elif con_type == 'tcp':
    cs = NonPersistentTCPCryptoService( conf )
  elif con_type == 'persistent_tcp':
    cs = PersistentTCPCryptoService( conf )
  else:
    raise ConfigurationError( f"unknown connection_type {con_type}" )
  return cs

##class CryptoService:
##
##  def __init__( self, conf ):
##    self.conf = conf
##    self.con_type = self.conf[ 'connectivity' ][ 'type' ]
##    if self.con_type == 'lib_cs':
##      self.cs = BaseCryptoService( self.conf )
##      BaseCryptoService.__init__( self,  self.conf )
##    elif self.con_type == 'tcp':
##      self.cs = NonPersistentTCPCryptoService( self.conf )
###      NonPersistentTCPCryptoService.__init__( self, self.conf )
##    else:
##      raise ConfigurationError( f"unknown connection_type {con_type}" )
##
##  def serve( self, req_bytes ) -> bytes :
##    if self.con_type == 'lib_cs':
##      return self.cs.serve( req_bytes )
##    else:
####      self.serve_forever()
##      ImplementationError( f"'serve' can only be used for con_type {self.con_type}."\
##                            "Current con_type is set to {self.con_type}")
##
##  def serve_forever( self ):
##    self.cs.serve_forver()
